//
//  **************************************************************************
//
//  controls.h 
//
//  (C) 2003 Bosco K. Ho 
//
//  **************************************************************************
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 2.1 of the License, or (at
//  your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,  but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details. 
//  
//  You should have received a copy of the GNU Lesser General Public License 
//  along with this program; if not, write to the Free Software Foundation, 
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//  **************************************************************************
//

#ifndef _controls_h
#define _controls_h

#include <string>
#include <vector>
#include <list>
#include <sstream>     
#include <iomanip>

#include "wx/dcbuffer.h"
#include "wx/wx.h"
#include "wx/spinctrl.h"

#include "controller.h"
#include "myprot.h"
#include "graphic3d.h"
#include "widget.h"
#include "protview.h"



// Displays the Ramachandran plot of selected residues

class RamaPlot : public Widget
{
public:
  RamaPlot(wxWindow* pParent, Controller* pController, int x, int y, int size):
    Widget(pParent, x, y, size, size), 
    mpController(pController),
    mIsLeftMouseDown(false) {}

  void onLeftMouseDrag();
  void onLeftMouseUp();
  void onLeftMouseDown();
  void onRightMouseDown();

  void draw(wxBufferedPaintDC &dc);
  
  int convertPhiToX(double phi);
  int convertPsiToY(double psi);
  
  void drawSlantedEllipse(wxBufferedPaintDC &dc, int x1, int y1, int x2, int y2, double width);
  
  Controller* mpController;
  bool mIsLeftMouseDown;
};



class RamaPlotSizer : public Widget
{
public:
  RamaPlotSizer(wxWindow* pParent, RamaPlot* pRamaPlot):
    Widget(pParent, -1, -1, 14, 14), 
    mpRamaPlot(pRamaPlot),
    mIsLeftMouseDown(false) {}

  int getX();
  int getY();
 
  void onLeftMouseDrag();
  void onLeftMouseUp();
  void onLeftMouseDown();

  void draw(wxBufferedPaintDC &dc);
  
  RamaPlot* mpRamaPlot;
  bool mIsLeftMouseDown;
};



class ChiBox : public Widget
{
public:
  ChiBox(wxWindow* pParent, Controller* pController, RamaPlot* pRamaPlot, int height):
    Widget(pParent, -1, -1, -1, height), 
    mpController(pController),
    mpRamaPlot(pRamaPlot),
    miChi(-1), mnChiMax(5)
    {}
 
  bool isOn();

  int getX() { return mpRamaPlot->getX(); }
  int getY() { return mpRamaPlot->getY() - getHeight() - 10; }
  int getWidth() { return mpRamaPlot->getWidth(); }
  
  void onLeftMouseDrag();
  void onLeftMouseUp();
  void onLeftMouseDown();
  void onRightMouseDown();

  int convertChiToX(double chi);
  double convertXToChi(int x);
  int convertIChiToY(int nChi);
  int convertYToIChi(int y);

  void draw(wxBufferedPaintDC &dc);
  
  Controller* mpController;
  RamaPlot* mpRamaPlot;
  int miChi;
  int mnChiMax;
};




// Displays the amino acid sequence of the protein

class Sequence : public Widget
{
public:
  Sequence(wxWindow* pParent, Controller* pController, int x, int y, int height):
    Widget(pParent, x, y, -1, height), 
    mResWidth(14),
    mResHeight(15),
    mpController(pController),
    mpProtein(pController->mpProtein) 
  { 
    checkBounds();
  }

  void onLeftMouseDrag();
  void onLeftMouseDClick();
  void onLeftMouseDown();
  void onRightMouseDown();

  void draw(wxBufferedPaintDC &dc);
  void shiftBounds(int i);
  
  int convertXToRes(int x);
  int convertResToX(int i);
  void checkBounds();

  bool mDragSelect;
  int mFixedLoopRes;
  int mLeftRes;
  int mRightRes;
  int mnDisplayRes;
  int mResWidth;
  int mResHeight;
  int mnDisplayResMax;
  
  Controller* mpController;
  MyProtein* mpProtein;
};  



// Displays the Ramachandran plot of selected residues

class FocusPicker : public Widget
{
public:
  FocusPicker(wxWindow* pParent, 
              Sequence* pSequence, 
              ProtView* pProtView, int y, int height):
      Widget(pParent, -1, y, -1, height), 
      mpController(pSequence->mpController),
      mpSequence(pSequence),
      mpProtView(pProtView) {}

  int getX();
  int getWidth();
  
  void onLeftMouseDrag();
  void onLeftMouseDown();

  void draw(wxBufferedPaintDC &dc);

protected:
  
  Controller* mpController;
  Sequence* mpSequence;
  ProtView* mpProtView;
};  



// Loop Solutions

class LoopSolutions : public Widget
{
public:
  LoopSolutions(wxWindow* pParent, Controller* pController, int x, int y, int height):
      Widget(pParent, x, y, -1, height), 
      mpController(pController) {}

  int getWidth();
  
  int convertXToSoln(int x);
  int convertSolnToX(int x);
  
  void onLeftMouseDown();
  void onLeftMouseDrag();

  void draw(wxBufferedPaintDC &dc);

  Controller* mpController;
};  



// Scrollbar

class Scrollbar : public FocusPicker
{
public:
  Scrollbar(wxWindow* pParent, Sequence* pSequence, 
            ProtView* pProtView, int y, int height):
            FocusPicker(pParent, pSequence, pProtView, y, height) {}

  void onLeftMouseDrag();
  void onLeftMouseDown();

  int convertXtoRes(int x);
  int convertResToX(int i);

  void draw(wxBufferedPaintDC &dc);
  int mOffset;
};  


// SlabBar

class SlabBar : public Widget
{
public:
  SlabBar(wxWindow* pParent, ProtView* pProtView, 
    int x, int y, int width, int height):
    Widget(pParent, x, y, width, height),
    mpProtView(pProtView),
    mMaxSlabThickness(15.0)
    {}

  void onLeftMouseDrag();
  void onLeftMouseDown();
  void draw(wxBufferedPaintDC &dc);
  
  ProtView* mpProtView;
  double mMaxSlabThickness;
};  


// PickControl

class PickControl : public Widget
{
public:
  PickControl(wxWindow* pParent, 
              ProtView* pProtView, int x, int y, int height);
  void onLeftMouseDown();
  void draw(wxBufferedPaintDC &dc);
  
  std::vector<std::string> labels;
  std::vector<int> lengths;
  ProtView* mpProtView;
  Controller* mpController;
};  


#endif
