//
//  ***************************************************************************
//
//  graphics.cpp
//  (C) 2004 Bosco K. Ho 
//
//  GraphicsCanvas handles Vector3d inputs, sorts by depth in 3d, and converts 
//  it into a 2d allegro bitmap. Requires camera information.
//
//  ***************************************************************************
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 2.1 of the License, or (at
//  your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,  but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details. 
//  
//  You should have received a copy of the GNU Lesser General Public License 
//  along with this program; if not, write to the Free Software Foundation, 
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//  ****************************************************************************
//

#include "graphic3d.h"
#include <math.h>

#define hypotXY(a, b) sqrt((a.x()-b.x())*(a.x()-b.x()) + (a.y()-b.y())*(a.y()-b.y()))
#define max(a,b) ((a) < (b) ? (b) : (a))



// graphical objects for display 
typedef enum 
{
  GFX_LINE,
  GFX_DASH_LINE,
  GFX_QUAD,
  GFX_SOLID_QUAD,
  GFX_CIRCLE,
  GFX_SOLID_CIRCLE,
  GFX_STRING,
};



void Project3d::initColorTable()
{
  mColorTable[RED] = wxColor(_T("RED"));
  mColorTable[BLUE] = wxColor(_T("BLUE"));
  mColorTable[GREEN] = wxColor(_T("GREEN"));
  mColorTable[ORANGE] = wxColor(_T("ORANGE"));
  mColorTable[LIGHT_ORANGE] = wxColor(_T("GOLD"));
  mColorTable[YELLOW] = wxColor(_T("YELLOW"));
  mColorTable[PURPLE] = wxColor(_T("PURPLE"));
  
  // set up a grey-scale in the top half of the palette
  for (int i = 128; i < 256; i++)
  {
     int c = (i-128)*2;
     mColorTable[i] = wxColor(c,c,c);
  }
}



void Project3d::setViewport(int x, int y, int w, int h)
{
   double scale = sqrt((double)h*h + w*w)/2.0;

   mPerspXScale = scale;
   mPerspYScale = scale;

   mXOffset = x + int(w/2);
   mYOffset = y + int(h/2);
}



// Calculates the matrix that transforms the camera position to the origin
// and the cameraFront direction to point in the positive z 

void Project3d::setCamera(Vector3d& pos, Vector3d& front, Vector3d& up)
{
   mCameraPos = pos;
   mCameraUp = up.normal();
   mCameraFront = front.normal();

   mProjectMatrix = SuperpositionMatrix(Vector3d(0, 0, 0), 
                                   Vector3d(0, -1, 0), 
                                   Vector3d(0, 0, 1),
                                   mCameraPos, 
                                   mCameraPos + mCameraUp, 
                                   mCameraPos + mCameraFront);
}



Vector3d Project3d::point(Vector3d p)
{
  Vector3d pOut = mProjectMatrix*p;

  pOut.x(pOut.x()/pOut.z()*mPerspXScale + mXOffset);
  pOut.y(pOut.y()/pOut.z()*mPerspYScale + mYOffset);
  
  return pOut;
}



bool Project3d::isXYOverSphere(Vector3d& v, double r, int screenX, int screenY, int *pZOut)
{
  Vector3d pOut = point(v);
  if (pOut.z() <= 1) 
    return false;
  *pZOut = int(pOut.z());

  Vector3d pWid = point(v + mCameraUp.scaled(r));
  if (pWid.z() <= 1) 
    return false;

  double rOut = hypotXY(pOut, pWid);

  return ( (screenX >= pOut.x()-rOut-1) && (screenX <= pOut.x()+rOut+1) &&
           (screenY >= pOut.y()-rOut-1) && (screenY <= pOut.y()+rOut+1) );
}



void Project3d::clear() 
{
  mGfxObjectList.clear();
}



inline bool comp(const GfxObject& a, const GfxObject& b) { return a.z > b.z; };



void Project3d::line(Vector3d& p1, Vector3d& p2, int color, double width) 
{
  Vector3d p1Out = point(p1);
  if (p1Out.z() <= 1.0) 
    return;

  Vector3d p2Out = point(p2);
  if (p2Out.z() <=  (1)) return;

  Vector3d p1Wid = point(p1 + mCameraUp.scaled(width));
  if (p1Wid.z() <=  1.0) 
    return;

  GfxObject gfxObject;
  gfxObject.id    = GFX_LINE;
  gfxObject.v[0]  = int(p1Out.x());
  gfxObject.v[1]  = int(p1Out.y());
  gfxObject.v[2]  = int(p2Out.x());
  gfxObject.v[3]  = int(p2Out.y());
  gfxObject.width = max(2, int(hypotXY(p1Out, p1Wid)));
  gfxObject.color = color;
  gfxObject.z   = (int)max(p1Out.z(), p2Out.z()); 

  mGfxObjectList.push_back(gfxObject);
}



void Project3d::dashLine(Vector3d& p1, Vector3d& p2, int color, double width) 
{
  line(p1, p2, color, width);
  GfxObject& g = mGfxObjectList.back();
  g.id = GFX_DASH_LINE;
}



void Project3d::solidLine(Vector3d& p1, Vector3d& p2, int color, double width) 
{
  Vector3d p1Out = point(p1);
  if (p1Out.z() <= 1.0) 
    return;

  Vector3d p1Wid = point(p1 + mCameraUp.scaled(width));
  if (p1Wid.z() <=  1.0) 
    return;

  double r1 = hypotXY(p1Out, p1Wid) + 1;

  Vector3d p2Out = point(p2);
  if (p2Out.z() <=  (1)) 
    return;

  Vector3d p2Wid = point(p2 + mCameraUp.scaled(width));
  if (p2Wid.z() <=  (1)) 
    return;

  double r2 = hypotXY(p2Out, p2Wid) + 1;
  
  double dx = p1Out.y() - p2Out.y();
  double dy = p2Out.x() - p1Out.x();

  double d  = sqrt(dx*dx + dy*dy);

  dx = dx / d;
  dy = dy / d;

  GfxObject gfxObject;

  gfxObject.id    = GFX_SOLID_QUAD;
  gfxObject.v[0]  = int(p1Out.x() + dx * r1);
  gfxObject.v[1]  = int(p1Out.y() + dy * r1);
  gfxObject.v[2]  = int(p2Out.x() + dx * r2);
  gfxObject.v[3]  = int(p2Out.y() + dy * r2);
  gfxObject.v[4]  = int(p2Out.x() - dx * r2);
  gfxObject.v[5]  = int(p2Out.y() - dy * r2);
  gfxObject.v[6]  = int(p1Out.x() - dx * r1);
  gfxObject.v[7]  = int(p1Out.y() - dy * r1);
  gfxObject.color = color;
  gfxObject.z     = (int)max(p1Out.z(), p2Out.z()); 

  mGfxObjectList.push_back(gfxObject);
}



void Project3d::solidQuad(Vector3d v[], int color)
{
  GfxObject gfxObject;

  gfxObject.color = color;
  gfxObject.id    = GFX_SOLID_QUAD;
  gfxObject.z     = 0;
  for (int i=0; i<=3; i+=1) 
  {
    Vector3d pOut = point(v[i]);
    if (pOut.z() <= (1)) 
      return;

    gfxObject.v[i*2]     = int(pOut.x());
    gfxObject.v[i*2 + 1] = int(pOut.y());
    if (int(pOut.z()) > gfxObject.z) 
      gfxObject.z = int(pOut.z());
  }

  mGfxObjectList.push_back(gfxObject);
}



void Project3d::quad(Vector3d v[], int color, double width)
{
  GfxObject gfxObject;

  gfxObject.color = color;
  gfxObject.id    = GFX_QUAD;
  gfxObject.z     = 0;
  for (int i=0; i<=3; i+=1) 
  {
    Vector3d pOut = point (v[i]);
    if (pOut.z() <= (1)) return;
    gfxObject.v[i*2]     = int(pOut.x());
    gfxObject.v[i*2 + 1] = int(pOut.y());
    if (int(pOut.z()) > gfxObject.z) 
      gfxObject.z = int(pOut.z());
  }

  Vector3d p1Wid = point(v[0] + mCameraUp.scaled(width));
  if (p1Wid.z() <=  1.0) 
    return;
    
  gfxObject.width = max(2, int(hypotXY(point(v[0]), p1Wid)));

  mGfxObjectList.push_back(gfxObject);
}



void Project3d::solidSphere(Vector3d& p, double r, int color)
{
  Vector3d pOut = point(p);
  if (pOut.z() <=  (1)) return;

  Vector3d pWid = point(p + mCameraUp.scaled(r));
  if (pWid.z() <= (1)) return;

  double diffX = pOut.x()-pWid.x();
  double diffY = pOut.y()-pWid.y();
  double rOut = sqrt(diffX*diffX + diffY*diffY);

  GfxObject gfxObject;

  gfxObject.id    = GFX_SOLID_CIRCLE;
  gfxObject.v[0]  = int(pOut.x());
  gfxObject.v[1]  = int(pOut.y());
  gfxObject.v[2]  = int(rOut) + 1;
  gfxObject.color = color;
  gfxObject.z     = int(pOut.z());

  mGfxObjectList.push_back(gfxObject);
}



void Project3d::ring(Vector3d& p, double r, double width, int color)
{
  Vector3d p1 = point(p);
  if (p1.z() <=  (1)) return;

  Vector3d p2 = point(p + mCameraUp.scaled(r));
  if (p2.z() <= (1)) return;

  Vector3d p3 = point(p + mCameraUp.scaled(r + width));
  if (p3.z() <= (1)) return;

  double rOut = hypotXY(p1, p2) + 1;

  GfxObject gfxObject;

  gfxObject.id    = GFX_CIRCLE;
  gfxObject.v[0]  = int(p1.x());
  gfxObject.v[1]  = int(p1.y());
  gfxObject.v[2]  = int(rOut);
  gfxObject.width = max(2, int(hypotXY(p2, p3)));
  gfxObject.color = color;
  gfxObject.z     = int(p1.z());

  mGfxObjectList.push_back(gfxObject);
}



void Project3d::string(Vector3d& p, std::string* pString, int color)
{
  Vector3d p1 = point(p);
  if (p1.z() <=  (1)) return;

  GfxObject gfxObject;

  gfxObject.id    = GFX_STRING;
  gfxObject.v[0]  = int(p1.x());
  gfxObject.v[1]  = int(p1.y());
  gfxObject.color = color;
  gfxObject.z     = int(p1.z());
  gfxObject.pString = pString;

  mGfxObjectList.push_back(gfxObject);
}



wxColor Project3d::getZColor(int index, int z)
{
   wxColour rawColor = mColorTable[index];
   double fraction = (z - mMinZ)/(double)mDeltaZ;
   if (fraction < 0.05)
     fraction = 0.05;
   if (fraction < 1)
     return wxColor((int)(fraction*rawColor.Red()), 
                    (int)(fraction*rawColor.Green()),
                    (int)(fraction*rawColor.Blue()));
   else
     return rawColor;
}                             
                             
                             
void Project3d::drawZ(wxBufferedPaintDC &dc) 
{
  wxPoint polygon[4];
  wxString text;

  GfxObjectIter i;
  for (i = mGfxObjectList.begin(); i != mGfxObjectList.end(); i++) 
  {
    switch (i->id) 
    {
      case GFX_LINE:
        dc.SetPen( wxPen(getZColor(i->color, i->z), i->width, wxSOLID) );
        dc.DrawLine( i->v[0], i->v[1], i->v[2], i->v[3] );
        break;
      case GFX_DASH_LINE:
        dc.SetPen( wxPen(getZColor(i->color, i->z), i->width, wxDOT) );
        dc.DrawLine( i->v[0], i->v[1], i->v[2], i->v[3] );
        break;
      case GFX_SOLID_CIRCLE:
        dc.SetPen( wxPen( getZColor(DARKEST_GREY, i->z), 1, wxSOLID) );
        dc.SetBrush( getZColor(i->color, i->z) );
        dc.DrawCircle(i->v[0], i->v[1], i->v[2]);
        break;
      case GFX_CIRCLE:
        dc.SetPen( wxPen(getZColor(i->color, i->z), i->width, wxSOLID) );
        dc.SetBrush( *wxTRANSPARENT_BRUSH );
        dc.DrawCircle(i->v[0], i->v[1], i->v[2]);
        break;
      case GFX_STRING:
        text = _T(i->pString->c_str());
        dc.SetFont(wxFont(10, wxMODERN, wxNORMAL, wxNORMAL));
        dc.SetBackgroundMode(wxTRANSPARENT);
        dc.SetTextForeground(*wxBLACK);
        dc.DrawText(text, i->v[0] - 1, i->v[1] - 1);    
        dc.DrawText(text, i->v[0] - 1, i->v[1] - 1);
        dc.DrawText(text, i->v[0]    , i->v[1] - 1);
        dc.DrawText(text, i->v[0] - 1, i->v[1]    );
        dc.DrawText(text, i->v[0] + 1, i->v[1] + 1);
        dc.DrawText(text, i->v[0]    , i->v[1] + 1);
        dc.DrawText(text, i->v[0] + 1, i->v[1]    );
        dc.DrawText(text, i->v[0] - 1, i->v[1] + 1);
        dc.DrawText(text, i->v[0] + 1, i->v[1] - 1);
        dc.SetTextForeground(getZColor(i->color, i->z));
        dc.DrawText(text, i->v[0]    , i->v[1]    );
        break;
      case GFX_QUAD:
        polygon[0] = wxPoint(i->v[0], i->v[1]);
        polygon[1] = wxPoint(i->v[2], i->v[3]);
        polygon[2] = wxPoint(i->v[4], i->v[5]);
        polygon[3] = wxPoint(i->v[6], i->v[7]);
        dc.SetBrush( *wxTRANSPARENT_BRUSH );
        dc.SetPen( wxPen(getZColor(i->color, i->z), i->width, wxSOLID) );
        dc.DrawPolygon(WXSIZEOF(polygon), polygon);
        break;
      case GFX_SOLID_QUAD:
        polygon[0] = wxPoint(i->v[0], i->v[1]);
        polygon[1] = wxPoint(i->v[2], i->v[3]);
        polygon[2] = wxPoint(i->v[4], i->v[5]);
        polygon[3] = wxPoint(i->v[6], i->v[7]);
        dc.SetPen( wxPen( getZColor(DARKEST_GREY, i->z), 1, wxSOLID) );
        dc.SetBrush( getZColor(i->color, i->z) );
        dc.DrawPolygon(WXSIZEOF(polygon), polygon);
        break;
    }
  }
}



void Project3d::drawZLight(wxBufferedPaintDC &dc) 
{
  mGfxObjectList.sort();

  mMinZ = mGfxObjectList.front().z;
  double maxZ = mGfxObjectList.back().z;
  mDeltaZ = maxZ - mMinZ;

  
  drawZ(dc); 
}



void Project3d::drawZFlat(wxBufferedPaintDC &dc) 
{
  // Exploits the choice made of making all things in the
  // closest 10% all equally bright
  // so we just artifically jig the minimum back so
  // that everything visible is in the closest 10%
  
  mGfxObjectList.sort();

  mMinZ = mGfxObjectList.front().z;
  double maxZ = mGfxObjectList.back().z;
  mDeltaZ = maxZ - mMinZ;
  mMinZ = mMinZ - 20.0*mDeltaZ;
  mDeltaZ = mDeltaZ / 2.0;
  
  drawZ(dc); 
}



void Project3d::drawZFixLight(wxBufferedPaintDC &dc, Vector3d lightCenter) 
{
  mGfxObjectList.sort();

  mMinZ = mGfxObjectList.front().z + 1.0;
  double maxZ = point(lightCenter).z() - 1.0;
  mDeltaZ = maxZ - mMinZ;
  
  drawZ(dc); 
}
