//
//  ***************************************************************************
//
//  loop.cpp
//  (C) 2004 Bosco K. Ho 
//
//
//  ***************************************************************************
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 2.1 of the License, or (at
//  your option) any later version.
//  
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details. 
//  
//  You should have received a copy of the GNU Lesser General Public License 
//  along with this program; if not, write to the Free Software Foundation, 
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//  ****************************************************************************
//

#include "loop.h"
#include "vector3d.h"


using namespace std;


void load_v_array(double v_array[], Vector3d& vec)
{
  v_array[0] = vec.x();
  v_array[1] = vec.y();
  v_array[2] = vec.z();
}



void load_vector3d(double v_array[], Vector3d& vec)
{
  vec.x(v_array[0]);
  vec.y(v_array[1]);
  vec.z(v_array[2]);
}



int LoopClosure::nSelectedRes()
{
  int result = 0;
  if (mRes1 >= 0)
    result += 1;
  if (mRes2 > mRes1)
    result += 1;
  if ((mRes3 > mRes2) && (mRes3 < mpProtein->nRes()))
    result += 1;
  return result;
}



void LoopClosure::setLoop(Protein* pProtein, int res1, int res2, int res3)
{
  mpProtein = pProtein;

  mRes1 = res1;
  mRes2 = res2;
  mRes3 = res3;
  
  mLoopRes1 = 0;
  mLoopRes2 = mRes2 - mRes1;
  mLoopRes3 = mRes3 - mRes1;
  
  if (nSelectedRes() == 3)
  {
    mpProtein->extractFragment(mLoop, mRes1, mRes3);
    extractParameters(&mLoop, mLoopRes1, mLoopRes2, mLoopRes3);  
    calculateSolutions();
  }
}


void LoopClosure::clear() 
{ 
  mRes1 = -1;
  mRes2 = -1;
  mRes3 = -1;
  clearSolns(mpSolnLoops); 
}


int LoopClosure::res(int i)
{
  if (i==0)
    return mRes1;
  if (i==1)
    return mRes2;
  if (i==2)
    return mRes3;
  return -1;
}


void LoopClosure::toggleRes(int res)
{
  if (res == mRes1)
  {
    mRes1 = mRes2;
    mRes2 = mRes3;
    mRes3 = -1;
  }  
  else if (res == mRes2)
  {
    mRes2 = mRes3;
    mRes3 = -1;
  }  
  else if (res == mRes3)
  {
    mRes3 = -1;
  }  
  else
  {  
    // new anchor point
    
    int n = nSelectedRes();

    if (n == 0)
      mRes1 = res;
    else if (n == 1)
    {
      if (res > mRes1)
        mRes2 = res;
      else 
      {
        mRes2 = mRes1;
        mRes1 = res;
      }
    }
    else if (n == 2)
    {
      if (res > mRes2)
        mRes3 = res;
      else if (res > mRes1)
      {
        mRes3 = mRes2;
        mRes2 = res;
      }
      else
      {
        mRes3 = mRes2;
        mRes2 = mRes1;
        mRes1 = res;
      }
    }
    else
    {
      if (res > mRes3)
        mRes3 = res;
      else if (res < mRes1)
        mRes1 = res;
      else
        mRes2 = res;
    }      

    if (nSelectedRes() == 3)
      setLoop(mpProtein, mRes1, mRes2, mRes3);
  }

}



bool LoopClosure::isSelectedRes(int res)
{
  return (mRes1 == res) || (mRes2 == res) || (mRes3 == res);
}



void LoopClosure::extractParameters(Protein* p, int res1, int res2, int res3)
{
  double bond_lengths[6], bond_angles[7], torsions[2];

  // bond lengths
  bond_lengths[0]= pos_distance(p->atom(res1, CA).pos, 
                                p->atom(res1, C).pos); 
  bond_lengths[1]= pos_distance(p->atom(res1, C).pos, 
                                p->atom(res2, N).pos); 
  bond_lengths[2]= pos_distance(p->atom(res2, N).pos, 
                                p->atom(res2, CA).pos); 
  bond_lengths[3]= pos_distance(p->atom(res2, CA).pos, 
                                p->atom(res2, C).pos); 
  bond_lengths[4]= pos_distance(p->atom(res2, C).pos, 
                                p->atom(res3, N).pos); 
  bond_lengths[5]= pos_distance(p->atom(res3, N).pos, 
                                p->atom(res3, CA).pos); 

  // bond angles
  bond_angles[0]= DEG2RAD*pos_angle(p->atom(res1, N).pos, 
                                    p->atom(res1, CA).pos, 
                                    p->atom(res1, C).pos); 
  bond_angles[1]= DEG2RAD*pos_angle(p->atom(res1, CA).pos, 
                                    p->atom(res1, C).pos, 
                                    p->atom(res2, N).pos); 
  bond_angles[2]= DEG2RAD*pos_angle(p->atom(res1, C).pos,  
                                    p->atom(res2, N).pos, 
                                    p->atom(res2, CA).pos); 
  bond_angles[3]= DEG2RAD*pos_angle(p->atom(res2, N).pos, 
                                    p->atom(res2, CA).pos, 
                                    p->atom(res2, C).pos); 
  bond_angles[4]= DEG2RAD*pos_angle(p->atom(res2, CA).pos, 
                                    p->atom(res2, C).pos, 
                                    p->atom(res3, N).pos); 
  bond_angles[5]= DEG2RAD*pos_angle(p->atom(res2, C).pos, 
                                    p->atom(res3, N).pos, 
                                    p->atom(res3, CA).pos); 
  bond_angles[6]= DEG2RAD*pos_angle(p->atom(res3, N).pos, 
                                    p->atom(res3, CA).pos, 
                                    p->atom(res3, C).pos); 
    
  // torsion angles between anchor residues
  torsions[0] = DEG2RAD*pos_dihedral(p->atom(res1, CA).pos, 
                                     p->atom(res1, C).pos, 
                                     p->atom(res2, N).pos,  
                                     p->atom(res2, CA).pos);
  torsions[1] = DEG2RAD*pos_dihedral(p->atom(res2, CA).pos, 
                                     p->atom(res2, C).pos, 
                                     p->atom(res3, N).pos,  
                                     p->atom(res3, CA).pos);

  // send parameters to the loop-closure library
  initialize_loop_closure(bond_lengths, bond_angles, torsions);
}



void LoopClosure::clearSolns(LoopSolns& solns)
{
  int n = (int)solns.size();
  for (int i=n-1; i >= 0; i--)
    delete solns[i];
  solns.clear();
}



bool LoopClosure::hasProlineViolation(Protein* p)
{
  bool result = false;
  for (int i=0; i < p->nRes(); i++)
    if (p->res(i).key == PRO)
      if (!p->projectProlineRing(i, p->phi(i)))
      {
        result = true;
        break;
      }
  return result;
}



/*
 * using the parameters extracted in extract_loop_closure_param, 
 * calculates the Solutions, which are placed in r_n_solns, 
 * r_ca_solns, r_c_solns 
 */
void LoopClosure::calculateSolutions()
{

  clearSolns(mpSolnLoops);

  double in_r_n1[3], in_r_ca1[3];
  load_v_array(in_r_n1, mpProtein->atom(mRes1, N).pos);
  load_v_array(in_r_ca1, mpProtein->atom(mRes1, CA).pos);

  double in_r_ca3[3], in_r_c3[3];
  load_v_array(in_r_ca3, mpProtein->atom(mRes3, CA).pos);
  load_v_array(in_r_c3, mpProtein->atom(mRes3, C).pos);

  int n_soln;
  double r_n_solns[MAX_LOOP_SOLN][3][3];
  double r_ca_solns[MAX_LOOP_SOLN][3][3];
  double r_c_solns[MAX_LOOP_SOLN][3][3];
  solve_3pep_poly(in_r_n1, in_r_ca1, in_r_ca3, in_r_c3, 
                  r_n_solns, r_ca_solns, r_c_solns, &n_soln);

  for (int k=0; k < n_soln; k++)
  {
    Vector3d new_ca2, new_c2, new_c1;

    load_vector3d(r_ca_solns[k][1], new_ca2);
    load_vector3d(r_c_solns[k][1], new_c2);
    load_vector3d(r_c_solns[k][0], new_c1);
     
    Protein* p_loop = new Protein(mLoop);
    rotateLoopToSoln(p_loop, new_ca2, new_c1, new_c2);

    if (!hasProlineViolation(p_loop))
      mpSolnLoops.push_back(p_loop);
    else
      delete p_loop;
  }
  
  miSolution = iSolutionClosestToAtom(mRes2, CA);
}



void LoopClosure::setSolution(int i)
{
  if ((i >= 0) && (i < nSolution())) 
  {
    copyLoopToProtein(mpSolnLoops[i], mRes1);
    mpProtein->extractAngles();
    miSolution = i;
  }
}



void LoopClosure::copySidechain(Protein* p_sidechain, int sc_res, Protein* p_target, int res)
{
  Matrix3d m = SuperpositionMatrix(
                 p_target->atom(res, C).pos, p_target->atom(res, CA).pos, p_target->atom(res, N).pos, 
                 p_sidechain->atom(sc_res, C).pos, p_sidechain->atom(sc_res, CA).pos, p_sidechain->atom(sc_res, N).pos);
  p_sidechain->transform(m);
  for (int i=0; i < p_sidechain->res(sc_res).nAtom(); i++)
  {
    AtomIter a = p_sidechain->res(sc_res).atomIters[i];
    int atomKey = a->key;
    if ((a->key >= CB) || (a->key == HA1))
      p_target->atom(res, atomKey).pos = a->pos;
  }
}



int LoopClosure::iSolutionClosestToAtom(int res, int atom_key)
{
  Vector3d& test_pos = mpProtein->atom(res, atom_key).pos;
  
  int loop_res = res - mRes1;
  int i_soln = 0;
  double min_dist_diff = 1E6;
  for (int k=0; k < nSolution(); k++)
  {
    Vector3d& soln_pos = mpSolnLoops[k]->atom(loop_res, atom_key).pos;
    double dist_diff = pos_distance(test_pos, soln_pos);

    if ((k==0) || (dist_diff < min_dist_diff))
    {
      min_dist_diff = dist_diff;
      i_soln = k;
    }
  }
  return i_soln;
}


void LoopClosure::copySolns(LoopSolns& saved_copy)
{
  clearSolns(saved_copy);
  int n_soln = nSolution();
  for (int i=0; i < n_soln; i++)
    saved_copy.push_back(new Protein(*(mpSolnLoops[i])));
}



void LoopClosure::restoreSolns(LoopSolns& saved_copy)
{
  clearSolns(mpSolnLoops);
  int n_saved_soln = (int)saved_copy.size();
  for (int i=0; i < n_saved_soln; i++)
    mpSolnLoops.push_back(new Protein(*saved_copy[i]));
  miSolution = iSolutionClosestToAtom(mRes2, CA);  
}



void LoopClosure::drive(int drive_res, double phi, double psi)
{
  if ((drive_res <= mRes1) || (drive_res == mRes2) || (drive_res >= mRes3))
    return;

  LoopSolns temp_soln_loops;
  copySolns(temp_soln_loops);

  int loop_drive_res = drive_res - mRes1;
  mLoop.setPhi(loop_drive_res, phi);
  mLoop.setPsi(loop_drive_res, psi);
  mLoopRes1 = 0;
  mLoopRes2 = mRes2 - mRes1;
  mLoopRes3 = mRes3 - mRes1;
  extractParameters(&mLoop, mLoopRes1, mLoopRes2, mLoopRes3);  
  calculateSolutions();

  if (nSolution() > 0)
  {
    setSolution(iSolutionClosestToAtom(drive_res, CA));
    clearSolns(temp_soln_loops);
  }
  else
  {
    restoreSolns(temp_soln_loops);
    temp_soln_loops.clear();
  }
}



void LoopClosure::rotateLoopToSoln(Protein* p_soln_loop, 
                       Vector3d new_ca2, Vector3d new_c1, Vector3d new_c2)
{
  Matrix3d rotation;
  
  mLoopRes1 = 0;
  mLoopRes2 = mRes2 - mRes1;
  mLoopRes3 = mRes3 - mRes1;

  // copy the new pivot atom (CA2)
  Vector3d old_ca2 = mLoop.atom(mLoopRes2, CA).pos;
  p_soln_loop->atom(mLoopRes2, CA).pos = new_ca2;


  // move segment CA1-CA2 in loop to match new CA2 position
  rotation = TwoPointSuperpositionMatrix(
               mpProtein->atom(mRes1, CA).pos, new_ca2,
               p_soln_loop->atom(mLoopRes1, CA).pos, old_ca2);
  for (int i=mLoopRes1; i<=mLoopRes2-1; i++)
    p_soln_loop->res(i).transform(rotation);
  if (p_soln_loop->hasAtom(mLoopRes2, H)) 
    p_soln_loop->atom(mLoopRes2, H).pos.transform(rotation);
  p_soln_loop->atom(mLoopRes2, N).pos.transform(rotation);


  // move segment CA2-CA3 in loop to match new CA2 position
  rotation = TwoPointSuperpositionMatrix(
               mpProtein->atom(mRes3, CA).pos, new_ca2, 
               p_soln_loop->atom(mLoopRes3, CA).pos, old_ca2);
  p_soln_loop->atom(mLoopRes2, C).pos.transform(rotation);
  p_soln_loop->atom(mLoopRes2, O).pos.transform(rotation);
  for (int i=mLoopRes2+1; i<=mLoopRes3; i++)
    p_soln_loop->res(i).transform(rotation);


  // rotate segment CA1-CA2 to match the new C1 atom
  rotation = SuperpositionMatrix(
              new_c1, mpProtein->atom(mRes1, CA).pos, new_ca2, 
              p_soln_loop->atom(mLoopRes1, C).pos, p_soln_loop->atom(mLoopRes1, CA).pos, new_ca2);
  p_soln_loop->atom(mLoopRes1, C).pos = new_c1;
  p_soln_loop->atom(mLoopRes1, O).pos.transform(rotation);
  for (int i=mLoopRes1+1; i<=mLoopRes2-1; i++)
    p_soln_loop->res(i).transform(rotation);
  if (p_soln_loop->hasAtom(mLoopRes2, H)) 
    p_soln_loop->atom(mLoopRes2, H).pos.transform(rotation);
  p_soln_loop->atom(mLoopRes2, N).pos.transform(rotation);


  // rotate segment CA2-CA3 to match the new C2 atom
  rotation = SuperpositionMatrix(
           new_c2, new_ca2, mpProtein->atom(mRes3, CA).pos, 
           p_soln_loop->atom(mLoopRes2, C).pos, new_ca2, p_soln_loop->atom(mLoopRes3, CA).pos);
  p_soln_loop->atom(mLoopRes2, C).pos = new_c2;
  p_soln_loop->atom(mLoopRes2, O).pos.transform(rotation);
  for (int i=mLoopRes2+1; i<=mLoopRes3-1; i++)
    p_soln_loop->res(i).transform(rotation);
  if (p_soln_loop->hasAtom(mLoopRes3, H)) 
    p_soln_loop->atom(mLoopRes3, H).pos.transform(rotation);
  p_soln_loop->atom(mLoopRes3, N).pos.transform(rotation);


  // as the mLoop may not align with the real protein
  // we must copy the non-changeable parts of the loop
  // directly from the protein
  if (p_soln_loop->hasAtom(mLoopRes1, H)) 
    p_soln_loop->atom(mLoopRes1, H).pos = mpProtein->atom(mRes1, H).pos;
  p_soln_loop->atom(mLoopRes1, N).pos = mpProtein->atom(mRes1, N).pos;
  p_soln_loop->atom(mLoopRes3, C).pos = mpProtein->atom(mRes3, C).pos;
  p_soln_loop->atom(mLoopRes3, O).pos = mpProtein->atom(mRes3, O).pos;

  
  // copy saved copy of sidechains to the hinges
  copySidechain(&mLoop, mLoopRes1, p_soln_loop, mLoopRes1);
  copySidechain(&mLoop, mLoopRes2, p_soln_loop, mLoopRes2);
  copySidechain(&mLoop, mLoopRes3, p_soln_loop, mLoopRes3);
}



void LoopClosure::copyLoopToProtein(Protein* p_loop, int res)
{
  int n_res = p_loop->nRes();
  for (int i=0; i < n_res; i++)
  {
    int j = i + res;
    for (int k=0; k < p_loop->res(i).nAtom(); k++)
    {
      AtomIter& a = p_loop->res(i).atomIters[k];
      mpProtein->res(j).atom(a->key).pos = a->pos;
    }
  }
}

