#include "myprot.h"

/*               
const double WEAK_HBOND_ENERGY_HIGH  = -1000;     // cal/mol
*/


void clear_bond_list(BondList& b)
{
  int size = b.size();
  for (int i=0; i<size; i++)
    delete b[i];
  b.clear();
}




//////////////////////////////////////////////
// Measure
//////////////////////////////////////////////


void Measure::clickOnAtom(AtomMarker a)
{
  std::vector<AtomMarker>::iterator i;
  
  bool isDelete = false;
  for (i = mAtomMarkerList.begin(); i != mAtomMarkerList.end(); i++)
    if (*i == a)
    {
      mAtomMarkerList.erase(i);
      isDelete = true;
      break;
    }
  if (!isDelete)
  {
    int n = mAtomMarkerList.size();
    if (n == mMaxMarker)
    {
      std::vector<AtomMarker>::iterator i;
      i = mAtomMarkerList.begin();
      mAtomMarkerList.erase(i);
    }
    mAtomMarkerList.push_back(a);
  }
}



std::string double_to_string(double d, int width, int precision)
{
  std::ostringstream ostr;
  if (width > 0)
    ostr << std::fixed << std::setw(width) << std::setprecision(precision) << d;     
  else
    ostr << std::fixed << std::setprecision(precision) << d;     
  return ostr.str();
}



void Measure::clear()
{
  mAtomMarkerList.clear();
}


int Measure::size()
{
  return mAtomMarkerList.size(); 
}



Vector3d& Measure::pos(int i)
{
  AtomMarker& marker = mAtomMarkerList[i];
  return marker.prot->atom(marker.res, marker.atom).pos;
}


Vector3d& Measure::caPos(int i)
{
  AtomMarker& marker = mAtomMarkerList[i];
  return marker.prot->atom(marker.res, CA).pos;
}



void Measure::clearLabel()
{
  int n = mLabelList.size();
  for (int i=0; i<n; i++)
    delete mLabelList[i];  
  mLabelList.clear();
}


void Measure::pushLabel(std::string s)
{
  mLabelList.push_back(new std::string(s));
}



///////////////////////////////////////////////////
// MyProtein




MyProtein::MyProtein(char chain_id, std::string name):
   Protein(chain_id, name)
{ 
  mSelect.reset();
  countCovalentBonds();
  countInteractions();

  mFocus.prot = this;
  mFocus.turnOff();
  mIsChanged = false;
  
  mLoopClosure.setLoop(this, -1, -1, -1);
}



MyProtein::MyProtein(const MyProtein& rhs)
{ 
  *this = rhs;
}



MyProtein& MyProtein::operator=(const MyProtein& rhs)
{
  clear();
  insertFragment(rhs, 0);
  mSelect = rhs.mSelect;

  countCovalentBonds();
  countInteractions();

  mFocus = rhs.mFocus;
  mFocus.prot = this;
  
  mLoopClosure.setLoop(this, mLoopClosure.res(0), mLoopClosure.res(1),
                       mLoopClosure.res(2));

  setPdbCode(rhs.getPdbCode());
  setId(rhs.getId());
  setName(rhs.getName());
  return *this;

}



// H-Bond routines

void MyProtein::geometricHbond(int i, int c, int o, int j, int n, 
                               double cutoff, BondList& bond_list)
{
  double no = pos_distance(atom(j, n).pos, atom(i, o).pos);
  double nc = pos_distance(atom(j, n).pos, atom(i, c).pos);

  if (nc < no) 
    return;
  
  if (no < cutoff)
    bond_list.push_back(new Bond(AtomMarker(this, i, o), AtomMarker(this, j, n)));
}



void MyProtein::dsspHbond(int i, int c, int o, int j, int n, int h,
                          double cutoff, 
                          BondList& bond_list, 
                          BondList& heavy_atom_bond_list)
{
  
  double ho = pos_distance(atom(j, h).pos, atom(i, o).pos);
  double hc = pos_distance(atom(j, h).pos, atom(i, c).pos);
  double no = pos_distance(atom(j, n).pos, atom(i, o).pos);
  double nc = pos_distance(atom(j, n).pos, atom(i, c).pos);

  // make sure it is a hydrogen bond, i.e. the h is between the donors
  if (nc < ho) 
    return;

  // the kabsch-sanders h-bond function!!!!
  static double Q_SQ = -27888.0; // 331.2 * 0.30 * 0.30
  double energy;
  
  // in cal/mol
  energy = (long)floor(Q_SQ/ho - Q_SQ/hc + Q_SQ/nc - Q_SQ/no + 0.5);

  //static double HBOND_ENERGY_HIGH  = -500;     // cal/mol
  if (energy < cutoff)
  {
    bond_list.push_back(new Bond(AtomMarker(this, i, o), AtomMarker(this, j, h)));
    heavy_atom_bond_list.push_back(new Bond(AtomMarker(this, i, o), AtomMarker(this, j, n)));
  }
}



void MyProtein::testHbond(int i, int c, int o, 
                          int j, int n, int h)
{
  static double CA_DISTANCE_SQ = 9.0;

  if (!hasAtom(i,CA)) return;
  if (!hasAtom(j,CA)) return;

  if ( pos_distance(atom(j, CA).pos, atom(i, CA).pos) 
        > CA_DISTANCE_SQ ) return;

  if (!hasAtom(i, c) || !hasAtom(i, o) ||
      !hasAtom(j, n))
    return;

  if (hasAtom(j, h))
    dsspHbond(i,c,o,j,n,h, -500.0, mWithHydrogenHbonds, mHeavyAtomHbonds);
  else
    geometricHbond(i,c,o,j,n, 3.5, mHeavyAtomHbonds);
}



void MyProtein::testHbondAcceptor(int i, int c, int o, int j)
{
  if (resKey(j) != PRO)
    testHbond(i, c, o, j, N, H);

  switch (resKey(j))
  {
  case SER:
  case THR:
    testHbond(i, c, o, j, OG1, HG11);
    break;
  case ARG:
    testHbond(i, c, o, j, NE2, HE11);
    testHbond(i, c, o, j, NH1, HH11);
    testHbond(i, c, o, j, NH1, HH12);
    testHbond(i, c, o, j, NH2, HH21);
    testHbond(i, c, o, j, NH2, HH21);
    break;
  case LYS:
    testHbond(i, c, o, j, NZ1, HZ11);
    testHbond(i, c, o, j, NZ1, HZ12);
    testHbond(i, c, o, j, NZ1, HZ13);
    break;
  case ASN:
    testHbond(i, c, o, j, ND2, HD21);
    testHbond(i, c, o, j, ND2, HD22);
    break;
  case GLN:
    testHbond(i, c, o, j, NE2, HE21);
    testHbond(i, c, o, j, NE2, HE22);
    break;
  case HIS:
    testHbond(i, c, o, j, ND1, HD11);
    testHbond(i, c, o, j, NE2, HE21);
    break;
  case TRP:
    testHbond(i, c, o, j, NE1, HE11);
    break;
  case TYR:
    testHbond(i, c, o, j, OH1, HH11);
    break;
  }
}



void MyProtein::countHbondsBetweenRes(int i, int j)
{
  testHbondAcceptor(i, C, O, j);

  switch (resKey(i))
  {
  case SER:
  case THR:
    testHbondAcceptor(i, CB, OG1, j);
    break;
  case ASP:
    testHbondAcceptor(i, CG1, OD1, j);
    testHbondAcceptor(i, CG1, OD2, j);
    break;
  case ASN:
    testHbondAcceptor(i, CG1, OD1, j);
    break;
  case GLU:
    testHbondAcceptor(i, CD1, OE1, j);
    testHbondAcceptor(i, CD1, OE2, j);
    break;
  case GLN:
    testHbondAcceptor(i, CD1, OE1, j);
    break;
  case TYR:
    testHbondAcceptor(i, CZ1, OH1, j);
    break;
  }
}



void MyProtein::countHBondsInRes(int i)
{
  switch (resKey(i))
  {
  case SER:
  case THR:
    testHbond(i, C, O, i, OG1, HG11);
    break;
  case ASP:
    testHbond(i, CG1, OD1, i, N, H);
    testHbond(i, CG1, OD2, i, N, H);
    break;
  case ASN:
    testHbond(i, CG1, OD1, i, N, H);
    testHbond(i, C, O, i, ND2, HD21);
    testHbond(i, C, O, i, ND2, HD22);
    break;
  case GLU:
    testHbond(i, CD1, OE1, i, N, H);
    testHbond(i, CD1, OE2, i, N, H);
    break;
  case GLN:
    testHbond(i, CD1, OE1, i, N, H);
    testHbond(i, C, O, i, NE2, HE21);
    testHbond(i, C, O, i, NE2, HE22);
    break;
  case TYR:
    testHbond(i, CZ1, OH1, i, N, H);
    testHbond(i, C, O, i, OH1, HH11);
    break;
    break;
  case ARG:
    testHbond(i, C, O, i, NE2, HE11);
    testHbond(i, C, O, i, NH1, HH11);
    testHbond(i, C, O, i, NH1, HH12);
    testHbond(i, C, O, i, NH2, HH21);
    testHbond(i, C, O, i, NH2, HH21);
    break;
  case LYS:
    testHbond(i, C, O, i, NZ1, HZ11);
    testHbond(i, C, O, i, NZ1, HZ12);
    testHbond(i, C, O, i, NZ1, HZ13);
    break;
  case HIS:
    testHbond(i, C, O, i, ND1, HD11);
    testHbond(i, C, O, i, NE2, HE21);
    break;
  case TRP:
    testHbond(i, C, O, i, NE1, HE11);
    break;
  }
}

 
 
void MyProtein::countHbonds()
{
  clear_bond_list(mWithHydrogenHbonds);
  clear_bond_list(mHeavyAtomHbonds);

  for (int i=0; i<nRes(); i++)
  {
    countHBondsInRes(i);
    for (int j=i+1; j<nRes(); j+=1)
    {
      countHbondsBetweenRes(j, i);
      if (j > i+1)
        countHbondsBetweenRes(i, j);
    }
  }
}



// Weak H-Bond routines

void MyProtein::testWeakHbond(int i, int c, int o, 
                              int j, int n, int h)
{
  static double CA_DISTANCE_SQ = 9.0;

  if (!hasAtom(i,CA)) return;
  if (!hasAtom(j,CA)) return;

  if ( pos_distance(atom(j, CA).pos, atom(i, CA).pos) > CA_DISTANCE_SQ ) return;

  if (!hasAtom(i, c) || !hasAtom(i, o) ||
      !hasAtom(j, n))
    return;

  if (hasAtom(j, h))
    dsspHbond(i,c,o,j,n,h, -1000.0, mWithHydrogenWeakHbonds, mHeavyAtomWeakHbonds);
  else
    geometricHbond(i,c,o,j,n, 3.6, mHeavyAtomWeakHbonds);
}



void MyProtein::testWeakHbondAcceptor(int i, int c, int o, int j)
{
  testWeakHbond(i, c, o, j, CA, HA1);

  if (resKey(j) == PRO) 
  {
    testWeakHbond(i, c, o, j, CD1, HD11);
    testWeakHbond(i, c, o, j, CD1, HD12);
  } 
  if (resKey(j) == GLY)
    testWeakHbond(i, c, o, j, CA, HA2);

}



void MyProtein::countWeakHbondsBetweenRes(int i, int j)
{
  // backbone-backbone weak H-bonds
  testWeakHbond(i, C, O, j, CA, HA1);
  if (resKey(i) == PRO)
    testWeakHbond(i, C, O, j, CD1, HD11);
  if (resKey(j) == GLY)
    testWeakHbond(i, C, O, j, CA, HA2);


  // backbone-sidechain weak H-bonds
  testWeakHbondAcceptor(i, C, O, j);

  // sidechain-acceptor weak H-bonds
  switch (resKey(i))
  {
  case SER:
  case THR:
    testWeakHbondAcceptor(i, CB, OG1, j);
    break;
  case ASP:
    testWeakHbondAcceptor(i, CG1, OD1, j);
    testWeakHbondAcceptor(i, CG1, OD2, j);
    break;
  case ASN:
    testWeakHbondAcceptor(i, CG1, OD1, j);
    break;
  case GLU:
    testWeakHbondAcceptor(i, CD1, OE1, j);
    testWeakHbondAcceptor(i, CD1, OE2, j);
    break;
  case GLN:
    testWeakHbondAcceptor(i, CD1, OE1, j);
    break;
  case TYR:
    testWeakHbondAcceptor(i, CZ1, OH1, j);
    break;
  }
}


 
void MyProtein::countWeakHbonds()
{
  clear_bond_list(mWithHydrogenWeakHbonds);
  clear_bond_list(mHeavyAtomWeakHbonds);

  for (int i=0; i<nRes(); i++)
  {
    for (int j=i+1; j<nRes(); j+=1)
    {
      countWeakHbondsBetweenRes(j, i);
      if (j > i+1)
        countWeakHbondsBetweenRes(i, j);
    }
  }
}



void MyProtein::countCovalentBonds()
{
  // clear mResBonds, careful to free all bonds
  for (unsigned int i=0; i<mResBonds.size(); i++)
    clear_bond_list(mResBonds[i]);
  mResBonds.clear();
  BondList b;
  mResBonds.insert(mResBonds.end(), nRes(), b);
    
  // count all the bonds within each residue
  for (int i=0; i<nRes(); i++)
    countAllCovalentBondsInRes(i);
  
  // count all the bonds between different residues
  clear_bond_list(mPeptideBonds);
  for (int i=1; i<nRes(); i++)
    countPeptideBond(i-1, C, i, N);
  
  countDisulfideBonds();
}



void MyProtein::countStericClashes()
{
  clear_bond_list(mStericClashes);

  int n_res = nRes();
  int i, i_atom, j, j_atom;
  for (i=0; i < n_res; i++)
  { 
    
    for (j=i; j < n_res; j++)
      if (pos_dist_squared(atom(i, CA).pos, atom(j, CA).pos) < 81.0)
      {
        int n_atom_i = res(i).nAtom();
        for (i_atom=0; i_atom<n_atom_i; i_atom++)
        {
          int n_atom_j = res(j).nAtom();
          for (j_atom=0; j_atom < n_atom_j; j_atom++)
          {
            Atom& atom_i = *(res(i).atomIters[i_atom]);
            Atom& atom_j = *(res(j).atomIters[j_atom]);

            bool check = true;

            if ((j-i) == 1)
            {
               // nearest neighbour clashes
               if ((atom_i.key == C) || (atom_i.key == O) || (atom_i.key == CA))
                 if ((atom_j.key == H) || (atom_j.key == N) || (atom_j.key == CA))
                    check = false;
               if ((atom_i.key == C) || (atom_j.key == CD1))
                 check = false;

            }
            else if (j==i)
            {
               // self-residue clashes
               if (resKey(i) == PRO)
                 check = false;
               else if (abs(atom_i.nesting - atom_j.nesting) <= 2)
                 check = false;
               else if ((atom_i.nesting < 0) && (atom_j.nesting < 0))
                 check = false;
               else 
               {
                 int nesting_limit = get_n_chi(resKey(i)) - 2;
                 if ((atom_i.nesting >= nesting_limit) && (atom_j.nesting >= nesting_limit))
                   check = false;
               }
                 
               // special check for Ramachandran Plot
               if ((atom_i.key == CB) && (atom_j.key == O))
                 check = true;
               if ((atom_i.key == O) && (atom_j.key == CB))
                 check = true;
               if ((atom_i.key == O) && (atom_j.key == H))
                 check = true;
               if ((atom_i.key == H) && (atom_j.key == O))
                 check = true;                 
            }
            
            if (check)
            {
              double cutoff = 0.9 * (atom_i.vdw + atom_j.vdw);

              // special backbone steric clashes
              if ((atom_i.key == N) && (atom_j.key == N))
                cutoff = 2.5;
             
              if (pos_dist_squared(atom_i.pos, atom_j.pos) < cutoff*cutoff)
                mStericClashes.push_back(new Bond(AtomMarker(this, i, atom_i.key),
                                                  AtomMarker(this, j, atom_j.key)));
            }
          }
        }
      }
  }
}



void MyProtein::countDisulfideBonds()
{
  static double DISULFIDE_DISTANCE_SQ = 3.0 * 3.0;    //  angstroms

  clear_bond_list(mDisulfideBonds);

  for (int i=0; i<nRes(); i++)
    if ((resKey(i) == CYS) && hasAtom(i, SG1))
    {
      for (int j=i+1; j<nRes(); j++)
      {
        if ((resKey(j) == CYS) && hasAtom(j, SG1))
          if ( pos_dist_squared(atom(j, SG1).pos, atom(i, SG1).pos)  
                < DISULFIDE_DISTANCE_SQ )
            mDisulfideBonds.push_back(new Bond(AtomMarker(this, i, SG1),
                                               AtomMarker(this, j, SG1)));
      }
   }
}



void MyProtein::countPeptideBond(int i, int a, int j, int b)
{
  if (hasAtom(i, a) && hasAtom(j, b))
    mPeptideBonds.push_back(new Bond(AtomMarker(this, i, a),
                                     AtomMarker(this, j, b)));
}



void MyProtein::countCovalentBondInRes(int i, int a, int b)
{
  if (hasAtom(i, a) && hasAtom(i, b))
    mResBonds[i].push_back(new Bond(AtomMarker(this, i, a),
                                    AtomMarker(this, i, b)));
}



void MyProtein::countAllCovalentBondsInRes(int i)
{
  // backbone bonds
  countCovalentBondInRes(i, CA, HA1);
  countCovalentBondInRes(i, CA,   C);
  countCovalentBondInRes(i,  C,   O);
  countCovalentBondInRes(i, CA,   N);
  countCovalentBondInRes(i,  N,   H);
  
  if (resKey(i) != GLY)
    countCovalentBondInRes(i, CA, CB);
  else
    countCovalentBondInRes(i, CA, HA2);
  
  // blocking groups
  if (resKey(i) == ACE)
  {
    countCovalentBondInRes(i, CA, H1);
    countCovalentBondInRes(i, CA, H2);
    countCovalentBondInRes(i, CA, H3);
  }
  else if (resKey(i) == NME)
  {
    countCovalentBondInRes(i, CA, H1);
    countCovalentBondInRes(i, CA, H2);
    countCovalentBondInRes(i, CA, H3);
  }
  else // charged terminii
  {
    countCovalentBondInRes(i, C,  OXT);
    countCovalentBondInRes(i, N,  H2);
    countCovalentBondInRes(i, N,  H3);
  }
  
  // sidechain bonds
  if (resKey(i) == PRO)
  {
    countCovalentBondInRes(i, CB,  CG1);
    countCovalentBondInRes(i, CG1, CD1);
    countCovalentBondInRes(i, CA,  N);
    countCovalentBondInRes(i, N,   CD1);
  
    countCovalentBondInRes(i, CD1, N);
    countCovalentBondInRes(i, CD1, HD11);
    countCovalentBondInRes(i, CD1, HD12);
    countCovalentBondInRes(i, CB,  HB1);
    countCovalentBondInRes(i, CB,  HB2);
    countCovalentBondInRes(i, CG1, HG11);
    countCovalentBondInRes(i, CG1, HG12);
  }
  
  // sidechain bonds
  
  switch (resKey(i))
  {
    case ALA:
      countCovalentBondInRes(i, CB, HB1);
      countCovalentBondInRes(i, CB, HB2);
      countCovalentBondInRes(i, CB, HB3);
      break;
    case CYS:
      countCovalentBondInRes(i, CB, HB1);
      countCovalentBondInRes(i, CB, HB2);
      countCovalentBondInRes(i, CB, SG1);
      countCovalentBondInRes(i, SG1, HG11);
      break;
    case SER:
      countCovalentBondInRes(i, CB,  HB1);
      countCovalentBondInRes(i, CB,  HB2);
      countCovalentBondInRes(i, CB,  OG1);
      countCovalentBondInRes(i, OG1, HG11);
      break;
    case ASP:
      countCovalentBondInRes(i, CB,  HB1);

      countCovalentBondInRes(i, CB,  HB2);
      countCovalentBondInRes(i, CB,  CG1);
      countCovalentBondInRes(i, CG1, OD1);
      countCovalentBondInRes(i, CG1, OD2);
      countCovalentBondInRes(i, OD1 ,HD11);
      countCovalentBondInRes(i, OD2 ,HD21);

      break;
    case ASN:
      countCovalentBondInRes(i, CB,  HB1);
      countCovalentBondInRes(i, CB,  HB2);
      countCovalentBondInRes(i, CB,  CG1);
      countCovalentBondInRes(i, CG1, OD1);
      countCovalentBondInRes(i, CG1, ND2);
      countCovalentBondInRes(i, ND2, HD21);
      countCovalentBondInRes(i, ND2, HD22);

      break;
    case GLU:
      countCovalentBondInRes(i, CB,  HB1);
      countCovalentBondInRes(i, CB,  HB2);
      countCovalentBondInRes(i, CB,  CG1);

      countCovalentBondInRes(i, CG1, HG11);
      countCovalentBondInRes(i, CG1, HG12);
      countCovalentBondInRes(i, CG1, CD1);
      countCovalentBondInRes(i, CD1, OE1);
      countCovalentBondInRes(i, CD1, OE2);
      countCovalentBondInRes(i, OE1 ,HE11);
      countCovalentBondInRes(i, OE2 ,HE21);

      break;
    case GLN:
      countCovalentBondInRes(i, CB,  HB1);
      countCovalentBondInRes(i, CB,  HB2);
      countCovalentBondInRes(i, CB,  CG1);
      countCovalentBondInRes(i, CG1, HG11);
      countCovalentBondInRes(i, CG1, HG12);
      countCovalentBondInRes(i, CG1, CD1);
      countCovalentBondInRes(i, CD1, OE1);
      countCovalentBondInRes(i, CD1, NE2);
      countCovalentBondInRes(i, NE2, HE21);
      countCovalentBondInRes(i, NE2, HE22);

      break;
    case THR:

      countCovalentBondInRes(i, CB,  HB1);

      countCovalentBondInRes(i, CB,  OG1);
      countCovalentBondInRes(i, OG1, HG11);

      countCovalentBondInRes(i, CB,  CG2);
      countCovalentBondInRes(i, CG2, HG21);
      countCovalentBondInRes(i, CG2, HG22);
      countCovalentBondInRes(i, CG2, HG23);

      break;
    case VAL:
      countCovalentBondInRes(i, CB,  HB1);

      countCovalentBondInRes(i, CB,  CG1);
      countCovalentBondInRes(i, CG1, HG11);
      countCovalentBondInRes(i, CG1, HG12);
      countCovalentBondInRes(i, CG1, HG13);

      countCovalentBondInRes(i, CB,  CG2);
      countCovalentBondInRes(i, CG2, HG21);
      countCovalentBondInRes(i, CG2, HG22);
      countCovalentBondInRes(i, CG2, HG23);

      break;
    case ILE:
      countCovalentBondInRes(i, CB,  HB1);

      countCovalentBondInRes(i, CB,  CG1);
      countCovalentBondInRes(i, CG1, HG11);
      countCovalentBondInRes(i, CG1, HG12);


      countCovalentBondInRes(i, CG1, CD1);
      countCovalentBondInRes(i, CD1, HD11);
      countCovalentBondInRes(i, CD1, HD12);
      countCovalentBondInRes(i, CD1, HD13);

      countCovalentBondInRes(i, CB,  CG2);
      countCovalentBondInRes(i, CG2, HG21);
      countCovalentBondInRes(i, CG2, HG22);
      countCovalentBondInRes(i, CG2, HG23);


      break;
    case LEU:
      countCovalentBondInRes(i, CB,  HB1);
      countCovalentBondInRes(i, CB,  HB2);

      countCovalentBondInRes(i, CB,  CG1);
      countCovalentBondInRes(i, CG1, HG11);

      countCovalentBondInRes(i, CG1, CD1);
      countCovalentBondInRes(i, CD1, HD11);
      countCovalentBondInRes(i, CD1, HD12);
      countCovalentBondInRes(i, CD1, HD13);
      countCovalentBondInRes(i, CG1, CD2);
      countCovalentBondInRes(i, CD2, HD21);
      countCovalentBondInRes(i, CD2, HD22);
      countCovalentBondInRes(i, CD2, HD23);

      break;
    case LYS:
      countCovalentBondInRes(i, CB,  HB1);
      countCovalentBondInRes(i, CB,  HB2);

      countCovalentBondInRes(i, CB,  CG1);
      countCovalentBondInRes(i, CG1, HG11);
      countCovalentBondInRes(i, CG1, HG12);

      countCovalentBondInRes(i, CG1, CD1);
      countCovalentBondInRes(i, CD1, HD11);
      countCovalentBondInRes(i, CD1, HD12);

      countCovalentBondInRes(i, CD1, CE1);
      countCovalentBondInRes(i, CE1, HE11);
      countCovalentBondInRes(i, CE1, HE12);

      countCovalentBondInRes(i, CE1, NZ1);
      countCovalentBondInRes(i, NZ1, HZ11);
      countCovalentBondInRes(i, NZ1, HZ12);
      countCovalentBondInRes(i, NZ1, HZ13);

      break;
    case MET:
      countCovalentBondInRes(i, CB, HB1);
      countCovalentBondInRes(i, CB, HB2);

      countCovalentBondInRes(i, CB, CG1);
      countCovalentBondInRes(i, CG1, HG11);
      countCovalentBondInRes(i, CG1, HG12);

      countCovalentBondInRes(i, CG1, SD1);

      countCovalentBondInRes(i, SD1, CE1);
      countCovalentBondInRes(i, CE1, HE11);
      countCovalentBondInRes(i, CE1, HE12);
      countCovalentBondInRes(i, CE1, HE13);


      break;
    case ARG:
      countCovalentBondInRes(i, CB, HB1);
      countCovalentBondInRes(i, CB, HB2);

      countCovalentBondInRes(i, CB, CG1);
      countCovalentBondInRes(i, CG1, HG11);
      countCovalentBondInRes(i, CG1, HG12);

      countCovalentBondInRes(i, CG1, CD1);
      countCovalentBondInRes(i, CD1, HD11);
      countCovalentBondInRes(i, CD1, HD12);

      countCovalentBondInRes(i, CD1, NE1);
      countCovalentBondInRes(i, NE1, HE11);

      countCovalentBondInRes(i, NE1, CZ1);

      countCovalentBondInRes(i, CZ1, NH1);
      countCovalentBondInRes(i, NH1, HH11);
      countCovalentBondInRes(i, NH1, HH12);

      countCovalentBondInRes(i, CZ1, NH2);
      countCovalentBondInRes(i, NH2, HH21);
      countCovalentBondInRes(i, NH2, HH22);
      break;
    case HIS:
      countCovalentBondInRes(i, CB, HB1);
      countCovalentBondInRes(i, CB, HB2);

      countCovalentBondInRes(i, CB, CG1);
      countCovalentBondInRes(i, CG1, ND1);
      countCovalentBondInRes(i, ND1, CE1);
      countCovalentBondInRes(i, CE1, NE2);
      countCovalentBondInRes(i, NE2, CD2);
      countCovalentBondInRes(i, CD2, CG1);

      countCovalentBondInRes(i, ND1, HD11);
      countCovalentBondInRes(i, CE1, HE11);
      countCovalentBondInRes(i, CG2, HG21);
      countCovalentBondInRes(i, CD2, HD21);
      countCovalentBondInRes(i, NE2, HE21);

      break;
    case TRP:
      countCovalentBondInRes(i, CB, HB1);
      countCovalentBondInRes(i, CB, HB2);

      countCovalentBondInRes(i, CB, CG1);
      countCovalentBondInRes(i, CG1, CD1);
      countCovalentBondInRes(i, CD1, NE1);
      countCovalentBondInRes(i, NE1, CE2);
      countCovalentBondInRes(i, CE2, CD2);
      countCovalentBondInRes(i, CD2, CG1);

      countCovalentBondInRes(i, CE2, CZ2);
      countCovalentBondInRes(i, CZ2, CH2);
      countCovalentBondInRes(i, CD2, CE3);
      countCovalentBondInRes(i, CE3, CZ3);
      countCovalentBondInRes(i, CZ3, CH2);

      countCovalentBondInRes(i, CD1, HD11);
      countCovalentBondInRes(i, NE1, HE11);
      countCovalentBondInRes(i, CG2, HG21);
      countCovalentBondInRes(i, CE3, HE31);
      countCovalentBondInRes(i, CZ2, HZ21);
      countCovalentBondInRes(i, CZ3, HZ31);
      countCovalentBondInRes(i, CH2, HH21);
      break;
    case PHE:
      countCovalentBondInRes(i, CB, HB1);

      countCovalentBondInRes(i, CB, HB2);

      countCovalentBondInRes(i, CB, CG1);
      countCovalentBondInRes(i, CG1, CD1);
      countCovalentBondInRes(i, CD1, CE1);
      countCovalentBondInRes(i, CE1, CZ1);
      countCovalentBondInRes(i, CZ1, CE2);
      countCovalentBondInRes(i, CE2, CD2);
      countCovalentBondInRes(i, CD2, CG1);

      countCovalentBondInRes(i, CD1, HD11);
      countCovalentBondInRes(i, CE1, HE11);
      countCovalentBondInRes(i, CZ1, HZ11);
      countCovalentBondInRes(i, CG2, HG21);
      countCovalentBondInRes(i, CD2, HD21);
      countCovalentBondInRes(i, CE2, HE21);

      break;
    case TYR:
      countCovalentBondInRes(i, CB, HB1);
      countCovalentBondInRes(i, CB, HB2);

      countCovalentBondInRes(i, CB, CG1);
      countCovalentBondInRes(i, CG1, CD1);
      countCovalentBondInRes(i, CD1, CE1);
      countCovalentBondInRes(i, CE1, CZ1);
      countCovalentBondInRes(i, CZ1, CE2);
      countCovalentBondInRes(i, CE2, CD2);
      countCovalentBondInRes(i, CD2, CG1);

      countCovalentBondInRes(i, CD1, HD11);
      countCovalentBondInRes(i, CE1, HE11);
      countCovalentBondInRes(i, CZ1, OH1);
      countCovalentBondInRes(i, OH1, HH11);
      countCovalentBondInRes(i, CG2, HG21);
      countCovalentBondInRes(i, CD2, HD21);
      countCovalentBondInRes(i, CE2, HE21);

      break;
  }
}



void MyProtein::makeMeasurements()
{
  int n = mMeasure.size();
  if (0 == n) return;

  Vector3d pos[4];
  for (int i = 0; i < n; i++)
    pos[i] = mMeasure.pos(i);

  mMeasure.clearLabel();
  if (n >= 2)
  {
    double dist = pos_distance(pos[0], pos[1]);
    mMeasure.pushLabel(double_to_string(dist, 0, 2));
  }

  if (n >= 3)
  {
    double ang = pos_angle(pos[0], pos[1], pos[2]);
    mMeasure.pushLabel(double_to_string(ang, 0, 2));
  }
  
  if (n == 3)
  {
    double dist = pos_distance(pos[1], pos[2]);
    mMeasure.pushLabel(double_to_string(dist, 0, 2));
  }
  else if (n >= 4)
  {
    double dihed = pos_dihedral(pos[0], pos[1], pos[2], pos[3]);
    mMeasure.pushLabel("(" + double_to_string(dihed, 0, 1) + ")");

    double ang = pos_angle(pos[1], pos[2], pos[3]);
    mMeasure.pushLabel(double_to_string(ang, 0, 1));

    double dist = pos_distance(pos[2], pos[3]);
    mMeasure.pushLabel(double_to_string(dist, 0, 1));
  }
  
}



void MyProtein::clearInteractions()
{
  clear_bond_list(mStericClashes);
  clear_bond_list(mDisulfideBonds);
  clear_bond_list(mWithHydrogenHbonds);
  clear_bond_list(mHeavyAtomHbonds);
  clear_bond_list(mWithHydrogenWeakHbonds);
  clear_bond_list(mHeavyAtomWeakHbonds);
}



void MyProtein::countInteractions()
{
  countStericClashes();
  countHbonds();
  countWeakHbonds();
  countDisulfideBonds();
}



void make_alphabet_protein(MyProtein* p_protein)
{  
  p_protein->clear();
  for (int i=0; i<20; i++)
  {
    p_protein->appendRes(ALA+i);
    p_protein->setPhi(i, -60);
    p_protein->setPsi(i, 120);
  } 
  p_protein->countCovalentBonds();
  p_protein->extractAngles();
  p_protein->countInteractions();
  p_protein->mSelect.reset();
  p_protein->setName("The Alphabet Protein");
}



void make_alanine_peptide(MyProtein* p_protein)
{  
  p_protein->clear();
  for (int i=0; i<12; i++)
  {
    p_protein->appendRes(ALA);
    p_protein->setPhi(i, -60);
    p_protein->setPsi(i, -40);
  } 
  p_protein->countCovalentBonds();
  p_protein->extractAngles();
  p_protein->countInteractions();
  p_protein->mSelect.reset();
  p_protein->setName("Alanine peptide");
}



