#ifndef _myprot_h_
#define _myprot_h_

#include "protein.h"
#include <deque>
#include <vector>
#include <sstream>     
#include <iomanip>
#include "loop.h"


//////////////////////////////////////////////
// AtomMarker
//////////////////////////////////////////////

class AtomMarker
{
public:
  AtomMarker(Protein* in_prot, int in_res, int in_atom):
    prot(in_prot), res(in_res), atom(in_atom) {};
  AtomMarker() {};

  Protein* prot;
  int res;
  int atom;

  bool operator == (const AtomMarker& a)
  { 
    return (prot == a.prot) && (res == a.res) && (atom == a.atom); 
  }

  Vector3d& pos() { return prot->atom(res, atom).pos; }
  char element() { return prot->atom(res, atom).element; }
  
  void turnOff() { res = -1; }

  void set(Protein* in_prot, int in_res, int in_atom)
  {
    prot = in_prot;
    res = in_res;
    atom = in_atom;
  }

  void set(AtomMarker& a)
  {
    prot = a.prot;
    res = a.res;
    atom = a.atom;
  }

  bool isValid() 
  { 
    if (res < 0)
      return false;
    else
      return prot->hasAtom(res, atom); 
  }
};



//////////////////////////////////////////////
// Bond
//////////////////////////////////////////////

class Bond
{
public:
  Bond(AtomMarker a1, AtomMarker a2):
    mAtom1(a1), mAtom2(a2) {};
  AtomMarker mAtom1;
  AtomMarker mAtom2;
};


typedef std::deque<Bond*> BondList;
void clear_bond_list(BondList& b);



//////////////////////////////////////////////
// SelectableList
//////////////////////////////////////////////


class SelectableList
{
public:

  void extend(int n)
  {
    int diff = n - (int)mSelectList.size();
    mSelectList.insert(mSelectList.end(), diff, false);
  }
  
  void set(int i, bool b) 
  { 
    if (i >= (int)mSelectList.size())
      extend(i+1);
    mSelectList[i] = b; 
  }
  
  bool isSelected(int i) 
  { 
    if (i >= (int)mSelectList.size())
    {
      extend(i+1);
      return false;
    }
    return mSelectList[i]; 
  }
  
  void reset() 
  { 
    mSelectList.clear();
  }

  int nSelected() 
  { 
    int size = 0; 
    for (int i=0; i < (int)mSelectList.size(); i++) 
      if (mSelectList[i])
        size += 1;
    return size; 
  }
  
  int orderedIndex(int order)
  {
    int curr_order = -1;
    for (int i=0; i < (int)mSelectList.size(); i++) 
    {
      if (mSelectList[i])
      {
        curr_order += 1;
        if (curr_order == order)
          return i;
      }
    }
    return -1;
  }        
        
  std::vector<bool> mSelectList;
};



//////////////////////////////////////////////
// Measure
//////////////////////////////////////////////


std::string double_to_string(double d, int width, int precision);

class Measure
{
public:
  Measure(): mMaxMarker(4) {};
  
  void clickOnAtom(AtomMarker a);
  void clear();
  int size();
  Vector3d& pos(int i);
  Vector3d& caPos(int i);
  void pushLabel(std::string s);
  void clearLabel();
   
  std::vector<std::string*> mLabelList;
  std::vector<AtomMarker> mAtomMarkerList;
  int mMaxMarker;
};




//////////////////////////////////////////////
// MyProtein
//////////////////////////////////////////////


class MyProtein : public Protein
{
public:
  MyProtein(char chain_id = ' ', std::string name = "untitled");
  MyProtein(const MyProtein& rhs);
  ~MyProtein() {}
  
  MyProtein& operator =(const MyProtein& rhs);

  void geometricHbond(int i, int c, int o, int j, int n, 
                      double cutoff, BondList& bond_list);
  void dsspHbond(int i, int c, int o, int j, int n, int h, 
                 double cutoff, BondList& bond_list, BondList& heavy_atom_bond_list);

  void testHbond(int i, int c, int o, int j, int n, int h);
  void testHbondAcceptor(int i, int c, int o, int j);
  void countHbondsBetweenRes(int i, int j);
  void countHBondsInRes(int i);
  void countHbonds();

  void testWeakHbond(int i, int c, int o, int j, int n, int h);
  void testWeakHbondAcceptor(int i, int c, int o, int j);
  void countWeakHbondsBetweenRes(int i, int j);
  void countWeakHbonds();
  
  void countPeptideBond(int i, int a, int j, int b);
  void countCovalentBondInRes(int i, int a, int b);
  void countAllCovalentBondsInRes(int i);
  void countCovalentBonds();

  void countStericClashes();
  
  void countDisulfideBonds();
  
  void makeMeasurements();
  
  void countInteractions();
  void clearInteractions();

  SelectableList mSelect;
  Measure mMeasure;
  LoopClosure mLoopClosure;

  std::vector<BondList> mResBonds;
  BondList mPeptideBonds;
  BondList mWithHydrogenHbonds;
  BondList mHeavyAtomHbonds;
  BondList mWithHydrogenWeakHbonds;
  BondList mHeavyAtomWeakHbonds;
  BondList mDisulfideBonds;
  BondList mStericClashes;
    
  AtomMarker mFocus;
  
  int mInsertRes;
  int mIsChanged;

};


void make_alphabet_protein(MyProtein* p_protein);
void make_alanine_peptide(MyProtein* p_protein);

#endif
