//
//  **************************************************************************
//
//  polymer.cpp
//
//  (C) 2003 Bosco K. Ho 
//
//  The purpose of the Polymer class is to provide an abstraction layer
//  to deal with residues. It's meant to be a base class for both
//  proteins and DNA. No build information or specific chemistry is
//  encoded in the structure. Some ideas adapted from the The Biomolecule
//  Toolkit (C) 2001, Eric Alm 
//
//  **************************************************************************
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 2.1 of the License, or (at
//  your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,  but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details. 
//  
//  You should have received a copy of the GNU Lesser General Public License 
//  along with this program; if not, write to the Free Software Foundation, 
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//  **************************************************************************
//

#include <iostream>
#include <iomanip>
#include <sstream>
#include <string>
#include <vector>
#include <list>
#include "polymer.h"
#include "residue.h"
#include "vector3d.h"

using namespace std;



ostream& operator << (ostream& os, const Atom& a)
{

  os << "Atom (atom name: " << a.name << ", " << a.key 
     << "  on: " << a.isOn << " built: " << endl;
  os << "  aa type: " << a.resName << ", " << a.resKey << " res no: " << a.resNo << endl;
  os << "  position: " << a.pos << ")" << endl;

  return os;
}




AtomIter Molecule::atomIter(int i) 
{ 
  AtomIter a = mAtomList.begin();
  for (int j=0; j<i; j++)
    a++;
  return a;
}



void Molecule::insertMolecule(int i, Molecule& molecule) 
{
  mAtomList.insert(atomIter(i), molecule.mAtomList.begin(), molecule.mAtomList.end());  
}



void Molecule::eraseAtoms(int i, int j) 
{
  mAtomList.erase(atomIter(i), atomIter(j));  
}



Molecule Molecule::extractMolecule(int i, int j) 
{
  Molecule fragment;
  fragment.mAtomList.insert(fragment.mAtomList.begin(), atomIter(i), atomIter(j));
  fragment.mId = mId;
  fragment.mName = mName;

  return fragment;
}



void Molecule::computeCenterOfMass() 
{
  double nAtom = 0;

  mCenterOfMass = Vector3d(0.0, 0.0, 0.0);
  AtomIter j = mAtomList.end();
  for (AtomIter i = mAtomList.begin(); i != j; i++)
  {
    mCenterOfMass += i->pos;
    nAtom += 1.0;
  }

  if (0.0 == nAtom) return;
  mCenterOfMass /= nAtom;
}



double Molecule::maxLength(void)
{
  double lengthSquared = 0.0;
  for (AtomIter i = mAtomList.begin(); i != mAtomList.end(); i++)
  {
    for (AtomIter j = i; j != mAtomList.end(); j++)
    {
      if (i != j)
      {
         double newLengthSquared = pos_dist_squared(i->pos, j->pos);
         if (newLengthSquared > lengthSquared)
           lengthSquared = newLengthSquared;
      }
    }
  }
  return sqrt(lengthSquared);
}



void Molecule::transform(const Matrix3d& matrix) 
{
  AtomIter j = mAtomList.end();
  for (AtomIter i = mAtomList.begin(); i != j; i++)
    (*i).pos  = matrix * ((*i).pos);
}



/////////////////////////
// Res


int Res::nAtom()
{
  return atomIters.size();
}



Atom& Res::atom(int atom_key)
{  
  AtomIterList::iterator a;
  for (a = atomIters.begin(); a != atomIters.end(); a++)
  {
    AtomIter& atom_iter = *a;
    if (atom_iter->key == atom_key)
      return *atom_iter;
  }
  return NULL_ATOM;
}



void Res::transform(Matrix3d matrix)
{
  AtomIterList::iterator a;
  for (a = atomIters.begin(); a != atomIters.end(); a++)
  {
    AtomIter& atom_iter = *a;
    atom_iter->pos.transform(matrix);
  }
}


/////////////////////////
// Polymer

// copy constructor
Polymer::Polymer(const Polymer& rhs): Molecule(rhs)
{
  clear();
  insertFragment(rhs, 0);
}



// -------- access functions

bool Polymer::hasAtom(int i, int atomKey) 
{
  if ( (i < 0) || (i > nRes() - 1) ) return false;

  for (int j = 0; j < nAtomInRes(i); j++)
    if (res(i).atomIters[j]->key == atomKey) 
      return true;

  return false;
}



int Polymer::nRes() const
{ 
  return mResList.size(); 
}



Atom& Polymer::atom(int i, int atomKey)
{
  for (int j = 0; j < res(i).nAtom(); j++)
  {
    if (res(i).atomIters[j]->key == atomKey)
      return *(res(i).atomIters[j]);
  }
  return NULL_ATOM;
}



void Polymer::checkAtomName(Atom& a)
{
  if (a.name == "")
    a.name = get_atom_name(a.resKey, a.key);
  a.element = a.name[1];

  if (a.element == 'H')
    a.vdw = 1.0;
  else if (a.name == " C  ")    
    a.vdw = 1.65;
  else if (a.element == 'C')
    a.vdw = 1.75;
  else if (a.element == 'O')
    a.vdw = 1.40;
  else if (a.element == 'N')
    a.vdw = 1.55;
  else if (a.element == 'S')
    a.vdw = 1.75;
  else
    a.vdw = 1.75;
    
  if (a.key == CB)
    a.nesting = -1;
  else if (a.key == CA)
    a.nesting = -2;
  else if ((a.key == HA1) || (a.key == N) || (a.key == C))
    a.nesting = -3;
  else if ((a.key == H) || (a.key == O))
    a.nesting = -4;
  else
  {
    a.nesting = -1;
    char c = a.name[2];
    if (isalpha(c))
    {
      std::string nesting_chars = "ABGDEZH";
      for (int k=0; k<7; k++)
        if (nesting_chars[k] == c)
        {
          a.nesting = k-2;
          if (a.element == 'H')
            a.nesting += 1;
          break;
        }
    }
  }
}



void Polymer::createAtom(int i, int atom_key, Vector3d pos)
{
  if (!hasAtom(i, atom_key))
  {
    Atom a;
    a.key  = atom_key; 
    a.name = get_atom_name(resKey(i), atom_key); 
    checkAtomName(a);
    Polymer::insertAtom(i, a);
  }
  atom(i, atom_key).pos = pos;
}



void Polymer::sproutAtom(int i, int a1, double length, 
      double angle, double dihed,int a2, int a3, int a4) 
{

  if (hasAtom(i, a2) && hasAtom(i, a3) && hasAtom(i, a4))
  {
    Vector3d pos = project_pos(length, angle, dihed,
                    atom(i, a2).pos, atom(i, a3).pos, atom(i, a4).pos);
    if (!hasAtom(i,a1)) 
      createAtom(i, a1, pos);
    else
      atom(i, a1).pos = pos;
  } 
}       



// constructs a string of the form 'LYS:32-CA'
string Polymer::resAtomName(int i, int a)
{
  ostringstream os;
  string aName = atom(i, a).name;
  while (' ' == aName[0]) aName.erase(0,1);
  while (' ' == aName[aName.length()-1]) aName.erase(aName.length()-1,aName.length());
  os << resName(i) << ":" << resNo(i) << "-" << aName; 
  return os.str(); 
}



double Polymer::maxLength(void)
{
  double lengthSquared = 0.0;
  double newLengthSquared;
  for (int i = 0; i < nRes(); i++)
    for (int ia = 0; ia < res(i).nAtom(); ia++)
      for (int j = i+1; j < nRes(); j++)
        for (int ja = 0; ja < res(j).nAtom(); ja++)
        {
          Vector3d& p1 = res(i).atomIters[ia]->pos;
          Vector3d& p2 = res(j).atomIters[ja]->pos;
          newLengthSquared = pos_dist_squared(p1, p2);
          if (newLengthSquared > lengthSquared)
            lengthSquared = newLengthSquared;
        }
  return sqrt(lengthSquared);
}



// -------- insertion/deletion of atoms 



// inserts atom into the molecule and updates residue block
void Polymer::insertAtom(int i, const Atom& atom) 
{
  Atom a = atom;
  a.resKey  = resKey(i); 
  a.resName = resName(i); 
  a.resNo   = resNo(i); 

  AtomIter insertPoint;
  if (i == nRes() - 1)
    insertPoint = lastAtomIter();
  else
    insertPoint = res(i+1).atomIters[0];
    // TODO: check if residue is empty
  
  AtomIter newAtomIter = mAtomList.insert(insertPoint, a);
  res(i).atomIters.push_back(newAtomIter);
}


  
// erase atom in the polymer, and updates residue block
void Polymer::eraseAtom(int i, int atomKey) 
{
  typedef AtomIterList::iterator AtomIterIter;
  
  AtomIterIter atom_iter_iter;
  AtomIterIter begin = res(i).atomIters.begin();
  AtomIterIter end = res(i).atomIters.end();

  for (atom_iter_iter = begin; atom_iter_iter != end; atom_iter_iter++) 
  {
     AtomIter atom_iter = *atom_iter_iter;
     if (atom_iter->key == atomKey) 
     {
       res(i).atomIters.erase(atom_iter_iter);
       mAtomList.erase(atom_iter);
       return;
     }
  }
}



// constructs the resBlk entries from the flat atom structures
// in the parent Molecule class
void Polymer::scanRes() 
{
  mResList.clear();

  int thisResNo = firstAtomIter()->resNo; 
  int i = -1;
  Res resBlk;
  for (AtomIter a = firstAtomIter(); a != lastAtomIter(); a++) 
  {
    // check when residue number changes
    if ( (a->resNo != thisResNo) || (i == -1) ) 
    { 
      thisResNo   = a->resNo;
      resBlk.no   = a->resNo;  
      resBlk.key  = a->resKey;  
      resBlk.name = a->resName;  

      mResList.push_back(resBlk);
      ++i;
    }
    res(i).atomIters.push_back(a);
  }
}



// inserts a residue at the position i,
// and moves old i to i+1
void Polymer::insertRes(int i, const Res& r)
{
  if ( (i < 0) || (i > nRes()) ) return;
  
  mResList.push_back(r); 

  // copies every residue from i to i+1 and updates resNo 
  for (int j = nRes() - 1; j > i; j--) 
  {
    res(j) = res(j-1);
    res(j).no += 1;
    for (int k = 0; k < res(j).nAtom(); k++)
      res(j).atomIters[k]->resNo += 1;
  }

  for (int k = 0; k < MAX_RES_PARAM; k++)
    res(i).param[k]  = r.param[k];

  if (nRes() == 1)
    res(i).no = 1;
  else
  {
    if (i == 0)
      res(i).no = res(i+1).no - 1;
    else
      res(i).no = res(i-1).no + 1;
  }
  res(i).key = r.key;
  res(i).name = r.name;
  res(i).atomIters.clear();

  for (unsigned int k = 0; k < r.atomIters.size(); k++)
  {
    Atom a = *(r.atomIters[k]);
    insertAtom(i, a);
  }

}



void Polymer::eraseRes(int i) 
{
  if (i > nRes() - 1) return;

  for (int j=0; j < nAtomInRes(i); j++)
  {
    int atom_key = res(i).atomIters[j]->key;
    eraseAtom(i, atom_key);
  }
  res(i).atomIters.clear();
  
  for (int j = i; j < nRes() - 1; j++) 
  {
    res(j) = res(j+1);      
    res(j).no -= 1;      
    for (int k = 0; k < nAtomInRes(j); k++)
      res(j).atomIters[k]->resNo -= 1;
  }

  mResList.pop_back();
} 



void Polymer::extractFragment(Polymer& p, int i, int j)
{
  p.clear();
  for (int k = j; k >= i; k--)
    p.insertRes(0, res(k));
}



void Polymer::insertFragment(const Polymer& p, int i)
{
  for (int j = p.nRes() - 1; j >= 0; j--)
    insertRes(i, p.mResList[j]);
}

