//
//  **************************************************************************
//
//  polymer.h
//
//  (C) 2003 Bosco K. Ho 
//
//  Adapted from the The Biomolecule Toolkit (C) 2001, Eric Alm 
//  
//  **************************************************************************
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 2.1 of the License, or (at
//  your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,  but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details. 
//  
//  You should have received a copy of the GNU Lesser General Public License 
//  along with this program; if not, write to the Free Software Foundation, 
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//  **************************************************************************
//

#ifndef polymer_h_
#define polymer_h_

#include <vector>
#include <iostream>
#include <string>
#include <list>
#include <map>

#include "vector3d.h"
#include "residue.h"


const double WATER_RADIUS      =  1.4;
const double DEFAULT_OCCUPANCY =  1.0;
const double DEFAULT_B_FACTOR  = -1.0;


/// Atom is a public data storage container
class Atom 
{
  public:
    Atom() {};
    
    Vector3d       pos;       // coordinates
    bool           isOn;
    char           chainId;
    int            no;       
    char           altLocNo; 
    std::string    name;  // name of atom types
    int            key;   // atom type
    int            resNo;     // residue number in the sequence 
    int            resKey;    // amino acid type
    std::string    resName;   // name of amino acid 
    char           insertCode;
    double         occup;
    double         bFactor;
    char           element;
    int            nesting;
    double         vdw;

};


std::ostream& operator << (std::ostream& os, const Atom& a);

static Atom NULL_ATOM;

typedef std::list<Atom>        AtomList;
typedef AtomList::size_type    AtomSizeType;
typedef AtomList::iterator     AtomIter;
typedef std::vector<AtomIter>  AtomIterList;

std::ostream& operator << (std::ostream& os, const Atom& a);



class Molecule 
{
public:
  Molecule::Molecule(char id = ' ', const std::string name = "untitled"): 
       mAtomList(0),  mId(id), mName(name), mCenterOfMass() {}
  Molecule(const Molecule& rhs):
       mAtomList(rhs.mAtomList), 
       mId(rhs.mId), 
       mName(rhs.mName), 
       mCenterOfMass(rhs.mCenterOfMass)  {}
  virtual ~Molecule() {};

  void setName(const std::string& name)  { mName = name; }
  std::string getName() const            { return mName; }
  
  void setId(char id)                    { mId = id; }
  char getId() const                     { return mId; }

  void insertMolecule(int i, Molecule& molecule);
  void eraseAtoms(int i, int j);
  Molecule extractMolecule(int i, int j);

  void computeCenterOfMass();
  Vector3d getCenterOfMass() const { return mCenterOfMass;}

  virtual double maxLength();
  
  void transform(const Matrix3d& matrix);

  AtomSizeType nAtom() const { return mAtomList.size(); }

  AtomIter atomIter(int i); 
  Atom& atom(int i)                          { return *atomIter(i);  }
  AtomIter firstAtomIter()                       { return mAtomList.begin(); }
  AtomIter lastAtomIter()                         { return mAtomList.end(); }
  void pushAtom(Atom& a)                     { mAtomList.push_back(a); }
  AtomIter insertAtom(AtomIter& iter, Atom& a)
                                             { return mAtomList.insert(iter, a); }
  void clear()                               { mAtomList.clear(); }

//protected:

  AtomList mAtomList;
  char mId;
  std::string mName;
  Vector3d mCenterOfMass;
};



const int MAX_RES_PARAM = 15;

// look-up table for blocks of atoms in a residue
class Res 
{
public:
  Res() {}
  int            key;             
  std::string    name;
  int            no;
  double         param[MAX_RES_PARAM];
  AtomIterList   atomIters;
  Atom&          atom(int atom_key);
  void           transform(Matrix3d matrix);
  int            nAtom();
};


class Polymer : public Molecule 
{
  public:
    Polymer(char chain_id = ' ', std::string name = "untitled") :
            Molecule(chain_id, name), mResList() {};
    Polymer(const Polymer& rhs);
    ~Polymer() {}

    void         clear() { mResList.clear();  Molecule::clear(); };
    bool         isEmpty() { return nRes() == 0; };
            
    Atom&        atom(int i, int atomKey);
    void         checkAtomName(Atom& a);
    void         sproutAtom(int i, int a1, double length, double angle, double dihed,
                             int a2, int a3, int a4);
    void         createAtom(int i, int atom_key, Vector3d pos); 
    bool         hasAtom(int i, int atomKey);
    void         insertAtom(int i, const Atom& a);
    void         eraseAtom(int i, int atomKey);

    void         scanRes();
    int          nRes() const;

    int          nAtomInRes(int i)      { return mResList[i].atomIters.size(); }

    Res&         res(int i) { return mResList[i]; }
    int          resKey(int i)          { return mResList[i].key; }
    int          resNo(int i)           { return mResList[i].no; }
    std::string  resName(int i)         { return mResList[i].name; }

    void         insertRes(int i, const Res& r);
    void         eraseRes(int i);
                 
    double       maxLength();
                 
    std::string  resAtomName(int i, int a);
                 
    void         extractFragment(Polymer& fragment, int i, int j);
    void         insertFragment(const Polymer& fragment, int i);
                 
    std::vector<Res> mResList;
};


#endif

