//
//  **************************************************************************
//
//  residue.h
//
//  (C) 2003 Bosco K. Ho 
//
//  name conversion routines for atom names and residue types
//
//  **************************************************************************
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 2.1 of the License, or (at
//  your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,  but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details. 
//  
//  You should have received a copy of the GNU Lesser General Public License 
//  along with this program; if not, write to the Free Software Foundation, 
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//  **************************************************************************
//

#ifndef residue_h_
#define residue_h_

#include <cstdio>
#include <vector>
#include <set>
#include <map>
#include <iostream>


/// enumerated type for amino acids
/// order follows alphabetical order of 1 letter code
const int N_RES_TYPE = 22;
typedef enum 
{
  XXX ,
  ALA , /* A */
  CYS , /* C */
  ASP , /* D */
  GLU , /* E */
  PHE , /* F */
  GLY , /* G */
  HIS , /* H */
  ILE , /* I */
  LYS , /* K */
  LEU , /* L */
  MET , /* M */
  ASN , /* N */
  PRO , /* P */
  GLN , /* Q */
  ARG , /* R */
  SER , /* S */
  THR , /* T */
  VAL , /* V */
  TRP , /* W */
  TYR , /* Y */
  ACE , /*   */
  NME   /*   */
};

int get_res_key(const std::string& s);
const char get_res_char(int key);
const std::string& get_res_name_3(int key);
const std::string& get_res_name(int key);


std::string get_atom_name(int resKey, int atomKey);
int get_atom_key(std::string s);
/// enumerated type for atom types the order goes from the CA up
/// so that varying an atom in a sidechain implies that he position of atoms with
/// a higher value also needs to be varied
const int N_ATOM_TYPE = 66 ;  
typedef enum 
{
  CA  ,
  C   ,
  HA1 ,
  N   ,
  O   ,
  H   ,
  H1  ,
  H2  ,
  H3  ,
  OXT ,
  CB  ,
  HA2 ,
  HB1 ,
  HB2 ,
  HB3 ,
  CG2 ,
  HG21,
  HG22,
  HG23,
  SG1 ,
  CG1 ,
  OG1 ,
  HG11,
  HG12,
  HG13,
  SD1 ,
  CD1 ,
  CD2 ,
  ND1 ,
  ND2 ,
  OD1 ,
  OD2 ,
  HD11,
  HD12,
  HD13,
  HD21,
  HD22,
  HD23,
  CE1 ,
  CE2 ,
  CE3 ,
  OE1 ,
  OE2 ,
  NE1 ,
  NE2 ,
  HE11,
  HE12,
  HE13,
  HE21,
  HE22,
  HE23,
  HE31,
  CZ1 ,
  CZ2 ,
  CZ3 ,
  NZ1 ,
  HZ11,
  HZ12,
  HZ13,
  HZ21,
  HZ31,
  CH2 ,
  OH1 ,
  NH1 ,
  NH2 ,
  HH11,
  HH12,
  HH21,
  HH22,   
  XX  
};



int get_n_chi(int resKey);
const int res_chi_tetrad[N_RES_TYPE+1][5][4] = 
{
  // XXX
  { 
    {  XX   }
  },
  // ALA
  { 
    {  XX    }
  },
  // CYS
  { 
    {  N, CA, CB, SG1 },
    { XX }
  },
  // ASP
  {  
    {  N, CA,  CB, CG1 },
    { CA, CB, CG1, OD1 },
    { XX }
  },
  // GLU
  {  
    {   N, CA, CB, CG1 },
    {  CA, CB, CG1, CD1 },
    {  CB, CG1, CD1, OE1 },
    {  XX }
  },
  // PHE
  {  
    {  N, CA,  CB, CG1 },
    { CA, CB, CG1, CD1 },
    { XX }
  },
  // GLY
  {  
    { XX }
  },
  // HIS
  {  
    {  N, CA,  CB, CG1 },
    { CA, CB, CG1, ND1 },
    { XX }
  },
  // ILE
  {  
    {  N, CA,  CB, CG1 },
    { CA, CB, CG1, CD1 },
    { XX }
  },
  // LYS
  {  
    {   N,  CA,  CB, CG1 },
    {  CA,  CB, CG1, CD1 },
    {  CB, CG1, CD1, CE1 },
    { CG1, CD1, CE1, NZ1 },
    {  XX }
  },
  // LEU
  {  
    {   N, CA,  CB, CG1 },
    {  CA, CB, CG1, CD1 },
    {  XX }
  },
  // MET
  {  
    {   N, CA,   CB, CG1 },
    {  CA, CB,  CG1, SD1 },
    {  CB, CG1, SD1, CE1 },
    {  XX }
  },
  // ASN
  {  
    {  N, CA,  CB, CG1 },
    { CA, CB, CG1, OD1 },
    { XX }
  },
  // PRO
  {  
    {  N,   CA,  CB,  CG1 },
    {  CA,  CB,  CG1, CD1 },
    {  CB,  CG1, CD1, N   },
    {  CG1, CD1, N,   CA  },
    { XX }
  },
  // GLN
  {  
    {   N,  CA,  CB, CG1 },
    {  CA,  CB, CG1, CD1 },
    {  CB, CG1, CD1, OE1 },
    {  XX }
  },
  // ARG
  {  
    {   N,   CA,  CB, CG1 },
    {  CA,   CB, CG1, CD1 },
    {  CB,  CG1, CD1, NE1 },
    {  CG1, CD1, NE1, CZ1 },
    {  XX }
  },
  // SER
  {  
    {  N, CA, CB, OG1 },
    { XX }
  },
  // THR
  {  
    {  N, CA, CB, OG1 },
    { XX }
  },
  // VAL
  {  
    {  N, CA, CB, CG1 },
    { XX }
  },
  // TRP
  {  
    {  N, CA,  CB, CG1 },
    { CA, CB, CG1, CD1 },
    { XX }
  },
  // TYR
  {  
    {  N, CA,  CB, CG1 },
    { CA, CB, CG1, CD1 },
    { XX }
  },
  // ACE
  {  
    { XX }
  },
  // NME
  {  
    { XX }
  },
};


#endif

