//
//  **************************************************************************
//
//  protein.cpp
//
//  (C) 2003 Bosco K. Ho 
//
//  The purpose of the Protein class is to provide a convenient 
//  structure to access the atoms via the Resblk structures without 
//  disturbing the underlying structure of the atoms.
//  The Res structure allows you to access the atom by 
//  residue number and atom-type. 
//
//  There are basically three categories of routines
//  
//  a) routines to read a pdb file and extract dihedral angles
//  b) routines to rotate and change internal parameters of a loaded
//     pdb file
//  c) routines to generate the residues de novo
//  d) various diagnostics to check protein integrity
//
//  In terms of de-novo strucutre determination, the information
//  to generate the structures are kept in the code instead of using
//  extensive data structures, I find this is easier to read and 
//  highlights the similarity between various amino acids.
//
//  Another important question is how do we index the residues and
//  chi angles? In C++, one assumes the count to start from 0 whereas
//  in most other cases, one starts counting from one.
//
//  Adapted from the The Biomolecule Toolkit (C) 2001, Eric Alm 
//
//  Cleaned the abstraction interface. Introduced a key for atoms
//  as well as for amino acids. This avoids the clumsy method of
//  using strings to identify atoms. There is too much confusion in the
//  peptide class since there is much interface between the protein
//  and peptide class. Consequently, the peptide class has been changed
//  to a lookup table class with little intrinsic functionality. 
//  Got rid of templates.
//
//
//  24/2/2004 fixed the HA1 and HA2 hydrogen sprout
//
//  **************************************************************************
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published
//  by the Free Software Foundation; either version 2.1 of the License, or (at
//  your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,  but
//  WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details. 
//  
//  You should have received a copy of the GNU Lesser General Public License 
//  along with this program; if not, write to the Free Software Foundation, 
//  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
//  **************************************************************************
//

#include <iostream>
#include <iomanip>
#include <sstream>
#include <string>
#include <vector>
#include <deque>
#include <list>
#include "protein.h"
#include "vector3d.h"

using namespace std;



// copy constructor
Protein::Protein(const Protein& rhs): 
  Polymer(rhs)
{
  mPdbCode = rhs.getPdbCode();
}



void Protein::extractAngles() 
{
  // extract backbone angles
  for (int i = 0; i < nRes(); i++) 
  {
    if (i > 0) 
    {
      phi(i)   = pos_dihedral(atom(i-1, C).pos, atom(i,  N).pos, 
                                 atom(i,  CA).pos, atom(i,  C).pos);
      omega(i) = pos_dihedral(atom(i-1,CA).pos, atom(i-1,C).pos, 
                                 atom(i,   N).pos, atom(i, CA).pos);
    }
    else 
    {
      if (resKey(i) == PRO)
      {
        if (hasAtom(i, CD1))
        {
          phi(i) = 180.0 
              + pos_dihedral(atom(i, CD1).pos, atom(i, N).pos, atom(i, CA).pos, atom(i, C).pos);
          normalize_angle(phi(i));
        }
        else
          phi(i) = 0;
      }
      else
      {
        if (hasAtom(i, H))
        {
          phi(i) = 180.0 
              + pos_dihedral(atom(i, H).pos, atom(i, N).pos, atom(i, CA).pos, atom(i, C).pos);
          normalize_angle(phi(i));
        }
        else
          phi(i)   = 0;
      }
      
      omega(i) = 180;
    }

    if (i < nRes() - 1) 
    {
      psi(i)   = pos_dihedral(atom(i,  N).pos, atom(i,  CA).pos, 
                                 atom(i,  C).pos, atom(i+1, N).pos);
    }
    else 
    {
      psi(i) = pos_dihedral(atom(i,  N).pos, atom(i,  CA).pos, 
                               atom(i,  C).pos, atom(i, O).pos) - 180;
      normalize_angle(psi(i));                         
    }
  }
  
  // extract sidechain angles  
  for (int i = 0; i < nRes(); i++) 
  {
    if (resKey(i) == PRO) 
    {
      extractProlineChiAngles(i);
      extractProlineParam(i);
    }
    else
    {
      for (int n = 0; n < get_n_chi(resKey(i)); n++) 
      {
        Vector3d p0 = atom(i, res_chi_tetrad[resKey(i)][n][0]).pos;
        Vector3d p1 = atom(i, res_chi_tetrad[resKey(i)][n][1]).pos;
        Vector3d p2 = atom(i, res_chi_tetrad[resKey(i)][n][2]).pos;
        Vector3d p3 = atom(i, res_chi_tetrad[resKey(i)][n][3]).pos;
  
        chi(i, n) = pos_dihedral(p0, p1, p2, p3); 
      }
    }
  }
}


void Protein::createRes(int i, int resKey) 
{
  Res r;

  r.key  = resKey;
  r.name = get_res_name_3(resKey);
  for (int j = 0; j < MAX_RES_PARAM; j++)
    r.param[j] = 180;
  r.atomIters.clear();

  Polymer::insertRes(i, r);

  if (1 == nRes())
    buildBackboneFromScratch(0);
  else if (0 == i)
    sproutBackboneFromNext(i);
  else 
    sproutBackboneFromPrev(i);
  sproutEnds(i);

  sproutWholeSidechain(i);
}



void Protein::appendRes(int key) 
{
  createRes(nRes(), key);
}



void Protein::mutateRes(int i, int new_res_key) 
{
  if (i > nRes() - 1) return;

  // erase sidechain atoms
  for (int j = res(i).nAtom() - 1; j >= 0; j--) 
  {
    int atom_key = res(i).atomIters[j]->key;
    if ((atom_key != C) && (atom_key != CA) && (atom_key != N))
      eraseAtom(i, atom_key);
  }

  res(i).key = new_res_key;
  res(i).name = get_res_name_3(new_res_key);
  for (int j = 0; j < nAtomInRes(i); j+=1)
  {
    res(i).atomIters[j]->resKey   = res(i).key;
    res(i).atomIters[j]->resName  = res(i).name;
  }

  sproutEnds(i);
  sproutWholeSidechain(i);
}



void Protein::splice(int target, int seg1, int seg2)
{
  // finds the transform so that the residues
  // seg1 is attached to the C-terminal of target
  // then transform seg1 to seg2 with that
  // transformation
  
  Vector3d a1, a2, a3; 
  a1 = atom(target, C).pos;

  // Project a2 = N(i+1) atom
  if (resKey(target) == ACE) 
  {
    Vector3d c_ca = atom(target, CA).pos - atom(target, C).pos;
    Vector3d ca_c  = atom(target, O).pos  - atom(target, C).pos;
    Vector3d axis = cross(c_ca, ca_c);
    Vector3d c_n = c_ca.rotated(-115.6, axis);
    a2 = atom(target, C).pos + c_n;
  }
  else
    // project N(i+1) atom
    a2 = project_pos(1.325, 115.0, psi(target),
            atom(target, N).pos, atom(target, CA).pos, a1);

  // project CA(i+1) atom
  a3 = project_pos(1.453, 121.0, 180.0, atom(target, CA).pos, a1, a2);
 
  Vector3d b1, b2, b3; 
  b3 = atom(seg1, CA).pos;
  b2 = atom(seg1, N).pos;

  // project C(i-1) atom
  if (resKey(seg1) == NME) 
  {
    Vector3d n_ca = atom(seg1, CA).pos - atom(seg1, N).pos;
    Vector3d n_h  = atom(seg1, H).pos  - atom(seg1, N).pos;
    Vector3d axis = cross(n_ca, n_h);
    Vector3d n_c = n_ca.rotated(-121.9, axis);
    b1 = atom(seg1, N).pos + n_c;
  }
  else
    b1  = project_pos(1.325, 121.9, phi(seg1), atom(seg1, C).pos, b3, b2);

  Matrix3d m = SuperpositionMatrix(a1, a2, a3, b1, b2, b3);
  
  for (int seg=seg1; seg <= seg2; seg++)
    res(seg).transform(m);
}



void Protein::eraseResAndSplice(int leftRes, int rightRes)
{
  if ( (leftRes > 0) && (rightRes < nRes() - 1) )
    splice(leftRes-1, rightRes+1, nRes() - 1);
  if (isEmpty()) 
    return;
  for (int j = leftRes; j <= rightRes; j++)
  {
    if (nRes() == 0) 
      return;
    eraseRes(leftRes);
  }
}



void Protein::insertFragmentAndSplice(Protein& fragment, int i)
{
  insertFragment(fragment, i); 
  if (i > 0)
    splice(i - 1, i, i + fragment.nRes() - 1); 
  if ( !isEmpty() && (i + fragment.nRes() < nRes()) )
    splice(i + fragment.nRes() - 1, i + fragment.nRes(), nRes() - 1); 
}  




// -------- de novo build routines



// standard covalent bond lengths
#define N_C           1.325
#define N_H           1.000
#define C_O           1.230
#define C_H           1.100
#define O_H           1.000
#define CA_C          1.530
#define CA_N          1.465
#define CD_OE         1.240
#define GLN_CD_OE     1.220
#define GLN_CD_NE     1.330
#define PHE_CZ_OH     1.360
#define CA_CB         1.530
#define CB_S          1.810
#define C_N           1.300

double PRO_CHI1      =  27.453;
double PRO_CHI5      =  -8.549;
double PRO_CA_CB_CG  =  103.726;
double PRO_CA_N_CD   =  111.560;
double PRO_N_CA_CB   =  103.655;
double PRO_CB_CG     =    1.501;
double PRO_CD_N      =    1.476;
double PRO_CA_CB     =    1.530;
double PRO_CA_N      =    1.465;
double PRO_CG_CD     =    1.513;


// standard covalent bond angles
#define C_C_S         112.7
#define CA_C_O        120.5
#define CA_N_H        115.0
#define CA_C_N        115.0
#define CA_N_C        121.0
#define C_CA_N        109.3
#define C_C_H         109.0
#define CA_CB_HB      117.0
#define CA_CB_CG      113.8



void Protein::buildBackboneFromScratch(int i) 
{
  createAtom(i, CA, Vector3d(0.0, 0.0, 1.2));

  Vector3d w(-1.0, 0.0, 0.0);

  Vector3d ca_n(0.0, 0.0, -1.0);
  ca_n.rotated (-C_CA_N * 0.5, w);
  ca_n.scaleTo (CA_N);

  createAtom(i, N, atom(i, CA).pos + ca_n);

  Vector3d ca_c = ca_n.rotated(+C_CA_N, w);
  ca_c.scaleTo(CA_C);
  createAtom(i, C, atom(i,CA).pos + ca_c);
}



void Protein::sproutBackboneFromPrev(int i) 
{
  if (resKey(i) == ACE) return;
  
  createAtom(i, N, project_pos(N_C,  CA_C_N, psi(i-1), 
                               atom(i-1, N).pos, atom(i-1, CA).pos, atom(i-1, C).pos));

  createAtom(i, CA, project_pos(CA_N, CA_N_C, omega(i), 
                                atom(i-1, CA).pos, atom(i-1, C).pos, atom(i, N).pos));
  
  if (resKey(i) == NME) return;

  createAtom(i, C, project_pos(CA_C, C_CA_N, phi(i), 
                               atom(i-1, C).pos, atom(i, N).pos, atom(i, CA).pos));
}

  
  
void Protein::sproutBackboneFromNext(int i) 
{
  if (resKey(i) == NME) return;
  
  createAtom(i, C, project_pos(N_C, CA_N_C, phi(i+1), 
                              atom(i+1, C).pos, atom(i+1, CA).pos, atom(i+1, N).pos));
  createAtom(i, CA, project_pos(CA_C, CA_C_N, omega(i+1), 
                                atom(i+1, CA).pos, atom(i+1, N).pos, atom(i, C).pos));
  
  if (resKey(i) == ACE) return;

  createAtom(i, N, project_pos(CA_N, C_CA_N, psi(i), 
                               atom(i+1, N).pos, atom(i, C).pos, atom(i, CA).pos));

}

  
  
void Protein::sproutEnds(int i) 
{
  if (resKey(i) == NME)
  {
    Vector3d c_n  = atom(i, N ).pos - atom(i-1, C).pos;
    Vector3d n_ca = atom(i, CA).pos - atom(i,   N).pos;
    Vector3d axis = cross(n_ca, c_n);
    Vector3d n_h  = n_ca.scaleTo(N_H).rotate(CA_N_H, axis);
    createAtom(i, H, atom(i, N).pos + n_h);
  }
  else if ((resKey(i) != ACE) && (resKey(i) != PRO))
    sproutAtom(i, H, N_H, CA_N_H, 180 + phi(i), C, CA, N);
  
  if (resKey(i) == ACE) 
  {
    Vector3d c_n   = atom(i+1, N ).pos - atom(i, C).pos;
    Vector3d c_ca  = atom(i,   CA).pos - atom(i, C).pos;
    Vector3d axis  = cross(c_n, c_ca);
    Vector3d ca_c   = c_ca.scaleTo(C_O).rotate(CA_C_O, axis);
    createAtom(i, O, atom(i, C).pos + ca_c);
  }
  else if (resKey(i) != NME)
     sproutAtom(i, O, C_O, CA_C_O, 180 + psi(i), N, CA, C);

  Vector3d ca_n = atom(i,N).pos - atom(i,CA).pos;
  Vector3d ca_c = atom(i,C).pos - atom(i,CA).pos;
  Vector3d v    = ca_n.scaleTo(1.0) + ca_c.scaleTo(1.0);
  Vector3d w    = ca_n.scaleTo(1.0) - ca_c.scaleTo(1.0);

  Vector3d ca_cb = v.rotated(125.5, w);

  if ((resKey(i) != ACE) && (resKey(i) != NME))
  {
    Vector3d ca_ha = v.rotated(-125.5, w).scaleTo(C_H);
    createAtom(i, HA1, atom(i,CA).pos + ca_ha);
  }

  if (resKey(i) == GLY)
  {
    createAtom(i, HA2, atom(i,CA).pos + ca_cb.scaleTo(C_H));
  }
  else if (resKey(i) == PRO) 
  {
    setBackbonePhi(i, PRO_CHI5 - 60);

    Vector3d ca_n = atom(i,N).pos - atom(i,CA).pos;
    Vector3d axis = cross(ca_n, ca_cb);
    ca_cb = ca_n.rotated(PRO_N_CA_CB, axis);
    createAtom(i, CB, atom(i, CA).pos + ca_cb.scaleTo(PRO_CA_CB));
    
    sproutAtom(i, CG1, PRO_CB_CG, PRO_CA_CB_CG, PRO_CHI1,  N, CA, CB);
    sproutAtom(i, CD1, PRO_CD_N,  PRO_CA_N_CD,  PRO_CHI5, CB, CA, N);
     
    createAtom(i, HD11, project_tetr_up_pos  (1.1, atom(i, N  ).pos, atom(i, CD1).pos, atom(i, CG1).pos)); 
    createAtom(i, HD12, project_tetr_down_pos(1.1, atom(i, N  ).pos, atom(i, CD1).pos, atom(i, CG1).pos)); 
    createAtom(i, HG11, project_tetr_up_pos  (1.1, atom(i, CD1).pos, atom(i, CG1).pos, atom(i, CB ).pos)); 
    createAtom(i, HG12, project_tetr_down_pos(1.1, atom(i, CD1).pos, atom(i, CG1).pos, atom(i, CB ).pos)); 
    createAtom(i, HB1 , project_tetr_up_pos  (1.1, atom(i, CG1).pos, atom(i, CB ).pos, atom(i, CA ).pos)); 
    createAtom(i, HB2 , project_tetr_down_pos(1.1, atom(i, CG1).pos, atom(i, CB ).pos, atom(i, CA ).pos)); 

    extractProlineChiAngles(i);
    extractProlineParam(i);
  }
  else if (resKey(i) == ACE) 
  {
    sproutAtom(i, H1, C_H, C_C_H,  60, O, C, CA);
    sproutAtom(i, H2, C_H, C_C_H, 180, O, C, CA);
    sproutAtom(i, H3, C_H, C_C_H, 300, O, C, CA);
  }
  else if (resKey(i) == NME) 
  {
    sproutAtom(i, H1, C_H, C_C_H,  60 + 60, H, N, CA);
    sproutAtom(i, H2, C_H, C_C_H, 180 + 60, H, N, CA);
    sproutAtom(i, H3, C_H, C_C_H, 300 + 60, H, N, CA);
  }
  else
  {
    createAtom(i, CB, atom(i,CA).pos + ca_cb.scaleTo(CA_CB));
  }

  if (resKey(i) == ALA) 
  {
    sproutAtom(i, HB1, C_H, C_C_H,  60, N, CA, CB);
    sproutAtom(i, HB2, C_H, C_C_H, 180, N, CA, CB);
    sproutAtom(i, HB3, C_H, C_C_H, 300, N, CA, CB);
  }
}



void Protein::sproutSidechain(int i, int n) 
{
  switch (n) 
  {
    case 0:
      // step one for chi(1)
      switch (resKey(i)) 
      {
        case ASP:
        case ASN:
        case HIS:
        case LEU:
        case GLN:
        case GLU:
        case ARG:
        case LYS:
        case TRP:
        case TYR:
        case PHE:
        case MET:
          sproutAtom(i,  CG1, CA_CB, 111.0, chi(i, n), N, CA, CB);
          break;
        case ILE:
          sproutAtom(i,  CG1, CA_CB, 111.0, chi(i, n),        N, CA, CB);
          sproutAtom(i,  HB1, C_H,   C_C_H, chi(i, n) + 120,  N, CA, CB);
          sproutAtom(i,  CG2, CA_CB, 111.0, chi(i, n) + 240,  N, CA, CB);
          break;
        case SER:
          sproutAtom(i,  OG1, 1.425, 112.0, chi(i, n),        N, CA, CB);
          sproutAtom(i,  HB1, C_H,   C_C_H, chi(i, n) + 120,  N, CA, CB);
          sproutAtom(i,  HB2, C_H,   C_C_H, chi(i, n) + 240,  N, CA, CB);

          sproutAtom(i, HG11, O_H, 110.0,                 180, CA, CB, OG1);
          break;
        case CYS:
          sproutAtom(i,  SG1, 1.830, 110.0, chi(i, n),  N, CA,  CB);
          sproutAtom(i, HG11, 1.350,  96.0,             0, CA, CB, SG1);
          break;
        case THR:
          sproutAtom(i,  OG1, 1.425, 104.0, chi(i, n),        N, CA, CB);
          sproutAtom(i,  HB1, C_H,   C_C_H, chi(i, n) + 120,  N, CA, CB);
          sproutAtom(i,  CG2, CA_CB, 111.0, chi(i, n) + 240,  N, CA, CB);
          sproutAtom(i, HG11, 1.000, 110.0, 180,                 CA, CB, OG1);
          break;
        case VAL:
          sproutAtom(i,  CG1, CA_CB, 111.0, chi(i, n),        N, CA, CB);
          sproutAtom(i,  HB1, C_H, C_C_H, chi(i, n) + 240,  N, CA, CB);
          sproutAtom(i,  CG2, CA_CB, 111.0, chi(i, n) + 120,  N, CA, CB);


          sproutAtom(i, HG11, C_H, C_C_H,  60, CA, CB, CG1);
          sproutAtom(i, HG12, C_H, C_C_H, 180, CA, CB, CG1);
          sproutAtom(i, HG13, C_H, C_C_H, 300, CA, CB, CG1);
          break;
      }
      // step two for chi(0)
      switch (resKey(i)) 
      {
        case ASP:
        case ASN:
        case HIS:
        case LEU:
        case GLN:
        case GLU:
        case LYS:
        case ARG:
        case TRP:
        case TYR:
        case PHE:
        case MET:
        case CYS:
          sproutAtom(i,  HB1, C_H, C_C_H, chi(i, n) + 120,  N, CA, CB);
          sproutAtom(i,  HB2, C_H, C_C_H, chi(i, n) + 240,  N, CA, CB);
          break;
        case THR:
        case ILE:
        case VAL:
          sproutAtom(i, HG21, C_H, C_C_H,  60, CA, CB, CG2);
          sproutAtom(i, HG22, C_H, C_C_H, 180, CA, CB, CG2);
          sproutAtom(i, HG23, C_H, C_C_H, 300, CA, CB, CG2);
          break;
      }
      break;
    case 1:
      // step one for chi(n)
      switch (resKey(i)) 
      {
        case ASN:
          sproutAtom(i,  OD1, 1.220, 120.0, chi(i, n)      , CA, CB, CG1);
          sproutAtom(i,  ND2, 1.330, 120.0, chi(i, n) + 180, CA, CB, CG1);

          sproutAtom(i, HD21, N_H, 120.0,    0, CB, CG1, ND2);
          sproutAtom(i, HD22, N_H, 120.0,  180, CB, CG1, ND2);
          break;
        case ASP:
          sproutAtom(i,  OD1, 1.220, 120.0, chi(i, n)      , CA, CB, CG1);
          sproutAtom(i,  OD2, 1.220, 120.0, chi(i, n) + 180, CA, CB, CG1);
          break;
        case ILE:
          sproutAtom(i,  CD1, CA_CB, 111.0, chi(i, n),       CA, CB, CG1);
          sproutAtom(i, HG11, C_H, C_C_H, chi(i, n) + 120, CA, CB, CG1);
          sproutAtom(i, HG12, C_H, C_C_H, chi(i, n) + 240, CA, CB, CG1);

          sproutAtom(i, HD11, C_H, C_C_H,   60, CB, CG1, CD1);

          sproutAtom(i, HD12, C_H, C_C_H,  180, CB, CG1, CD1);
          sproutAtom(i, HD13, C_H, C_C_H,  300, CB, CG1, CD1);
          break;
        case LEU:
          sproutAtom(i,  CD1, CA_CB, 111.0, chi(i, n),       CA, CB, CG1);
          sproutAtom(i,  CD2, CA_CB, 111.0, chi(i, n) + 120, CA, CB, CG1);
          sproutAtom(i, HG11, C_H, C_C_H, chi(i, n) + 240, CA, CB, CG1);

          sproutAtom(i, HD11, C_H, C_C_H,   60, CB, CG1, CD1);
          sproutAtom(i, HD12, C_H, C_C_H,  180, CB, CG1, CD1);
          sproutAtom(i, HD13, C_H, C_C_H,  300, CB, CG1, CD1);

          sproutAtom(i, HD21, C_H, C_C_H,   60, CB, CG1, CD2);
          sproutAtom(i, HD22, C_H, C_C_H,  180, CB, CG1, CD2);
          sproutAtom(i, HD23, C_H, C_C_H,  300, CB, CG1, CD2);
          break;
        case GLN:
        case GLU:
        case LYS:
        case ARG:
          sproutAtom(i,  CD1, CA_CB, 111.0, chi(i, n),       CA, CB, CG1);
          sproutAtom(i, HG11, C_H, C_C_H, chi(i, n) + 120, CA, CB, CG1);
          sproutAtom(i, HG12, C_H, C_C_H, chi(i, n) + 240, CA, CB, CG1);
          break;
        case MET:

          sproutAtom(i,  SD1, 1.530, C_C_H, chi(i, n),       CA, CB, CG1);
          sproutAtom(i, HG11, C_H, C_C_H, chi(i, n) + 120, CA, CB, CG1);
          sproutAtom(i, HG12, C_H, C_C_H, chi(i, n) + 240, CA, CB, CG1);
          break;
        case TRP:
          sproutAtom(i,  CD1, 1.340,  128.0,   chi(i, n),   CA, CB, CG1);
          sproutAtom(i,  CD2, 1.443, -126.6,   chi(i, n),   CA, CB, CG1);

          sproutAtom(i,  NE1, 1.380,  111.5, 180,   CB, CG1, CD1);
          sproutAtom(i,  CE2, 1.390,  107.4,   0,  CG1, CD1, NE1);

          sproutAtom(i,  CE3, 1.410,  129.6,   0,   CB, CG1, CD2);
          sproutAtom(i,  CZ3, 1.400,  115.4, 180,  CG1, CD2, CE3);
          sproutAtom(i,  CH2, 1.390,  124.6,   0,  CD2, CE3, CZ3);
          sproutAtom(i,  CZ2, 1.400,  129.6, 180,  CD1, NE1, CE2);

          sproutAtom(i, HD11, C_H,  124.3,   0,   CB, CG1, CD1);
          sproutAtom(i, HE11, N_H,  126.8, 180,  CG1, CD1, NE1);
          sproutAtom(i, HZ21, C_H,  125.8, 180,  CD2, CE2, CZ2);
          sproutAtom(i, HH21, C_H,  125.8, 180,  CE2, CZ2, CH2);
          sproutAtom(i, HZ31, C_H,  125.8, 180,  CD2, CE3, CZ3);
          sproutAtom(i, HE31, C_H,  125.8,   0,  CG1, CD2, CE3);
          break;
        case HIS:
          sproutAtom(i,  ND1, 1.384,  122.8,   chi(i, n),   CA, CB, CG1);
          sproutAtom(i,  CE1, 1.325,  108.4,   180,   CB, CG1, ND1);
          sproutAtom(i,  NE2, 1.364,  108.4,     0,  CG1, ND1, CE1);
          sproutAtom(i,  CD2, 1.530,  108.4,     0,  ND1, CE1, NE2);

          sproutAtom(i, HD11, C_H,  125.8,     0,   CB, CG1, ND1);
          sproutAtom(i, HE11, N_H,  125.8,   180,  CG1, ND1, CE1);
          sproutAtom(i, HD21, C_H,  125.8,     0,   CB, CG1, CD2);
          sproutAtom(i, HE21, N_H,  125.8,   180,  ND1, CE1, NE2);
          break; 
        case TYR:
        case PHE:
          sproutAtom(i,  CD1, 1.360, 120.0,  chi(i, n),  CA,  CB, CG1);
          sproutAtom(i,  CD2, 1.360,-120.0,  chi(i, n),  CA,  CB, CG1);
          sproutAtom(i,  CE1, 1.420, 120.0,  180,  CB, CG1, CD1);
          sproutAtom(i,  CE2, 1.420, 120.0,  180,  CB, CG1, CD2);
          sproutAtom(i,  CZ1, 1.360, 120.0,    0, CG1, CD1, CE1);

          sproutAtom(i, HD11, C_H, 120.0,    0,  CB, CG1, CD1);
          sproutAtom(i, HD21, C_H, 120.0,    0,  CB, CG1, CD2);

          sproutAtom(i, HE11, C_H, 120.0,  180, CG1, CD1, CE1);
          sproutAtom(i, HE21, C_H, 120.0,  180, CG1, CD2, CE2);

          if (resKey(i) == PHE)
            sproutAtom(i, HZ11, C_H, 120.0,  180, CD1, CE1, CZ1);
          else 
          {
            sproutAtom(i,  OH1, PHE_CZ_OH, 120.0,  180, CD1, CE1, CZ1);
            sproutAtom(i, HH11, C_H, 110.0,  180, CE1, CZ1, OH1);
          }

          break;
      }
      break;
    case 2:
      switch (resKey(i)) 
      {
        case GLN:
          sproutAtom(i,  OE1, GLN_CD_OE, 120.0, chi(i, n)      , CB, CG1, CD1);
          sproutAtom(i,  NE2, GLN_CD_NE, 120.0, chi(i, n) + 180, CB, CG1, CD1);

          sproutAtom(i, HE21, N_H, 120.0,    0, CG1, CD1, NE2);
          sproutAtom(i, HE22, N_H, 120.0,  180, CG1, CD1, NE2);
          break;
        case GLU:
          sproutAtom(i,  OE1, CD_OE, 120.0, chi(i, n)      , CB, CG1, CD1);
          sproutAtom(i,  OE2, CD_OE, 120.0, chi(i, n) + 180, CB, CG1, CD1);
          break;
        case ARG:
          sproutAtom(i,  NE1, CA_CB, C_C_H, chi(i, n)      , CB, CG1, CD1);
          sproutAtom(i, HD11, C_H, C_C_H, chi(i, n) + 120, CB, CG1, CD1);
          sproutAtom(i, HD12, C_H, C_C_H, chi(i, n) + 240, CB, CG1, CD1);
          break;
        case LYS:
          sproutAtom(i,  CE1, CA_CB, C_C_H, chi(i, n)      , CB, CG1, CD1);
          sproutAtom(i, HD11, C_H,   C_C_H, chi(i, n) + 120, CB, CG1, CD1);
          sproutAtom(i, HD12, C_H,   C_C_H, chi(i, n) + 240, CB, CG1, CD1);
          break;
        case MET:
          sproutAtom(i,  CE1, 1.530, C_C_H, chi(i, n)      , CB, CG1, SD1);
          sproutAtom(i, HD11, C_H, C_C_H, chi(i, n) + 120, CB, CG1, SD1);
          sproutAtom(i, HD12, C_H, C_C_H, chi(i, n) + 240, CB, CG1, SD1);

          sproutAtom(i, HE11, C_H, C_C_H,  60, CG1, SD1, CE1);
          sproutAtom(i, HE12, C_H, C_C_H, 180, CG1, SD1, CE1);
          sproutAtom(i, HE13, C_H, C_C_H, 300, CG1, SD1, CE1);
          break;
      }
      break;
    case 3:
      switch (resKey(i)) 
      {
        case ARG:
          sproutAtom(i,  CZ1, 1.530, 120.0, chi(i, n), CG1, CD1, NE1);
          sproutAtom(i, HE11, C_H, 240.0, chi(i, n), CG1, CD1, NE1);

          sproutAtom(i,  NH1, 1.530, 120.0,   0,  CD1, NE1, CZ1);
          sproutAtom(i,  NH2, 1.530, 120.0, 180,  CD1, NE1, CZ1);

          sproutAtom(i, HH11, N_H, 120.0,   0,  NE1, CZ1, NH1);
          sproutAtom(i, HH12, N_H, 120.0, 180,  NE1, CZ1, NH1);

          sproutAtom(i, HH21, N_H, 120.0,   0,  NE1, CZ1, NH2);
          sproutAtom(i, HH22, N_H, 120.0, 180,  NE1, CZ1, NH2);

          break;
        case LYS:
          sproutAtom(i,  NZ1, 1.530, 111.0, chi(i, n),       CG1, CD1, CE1);
          sproutAtom(i, HE11,   C_H, C_C_H, chi(i, n) + 120, CG1, CD1, CE1);
          sproutAtom(i, HE12,   C_H, C_C_H, chi(i, n) + 240, CG1, CD1, CE1);

          sproutAtom(i, HZ11,   C_H, C_C_H,   60, CD1, CE1, NZ1);
          sproutAtom(i, HZ12,   C_H, C_C_H,  180, CD1, CE1, NZ1);
          sproutAtom(i, HZ13,   C_H, C_C_H,  300, CD1, CE1, NZ1);
          break;
      }
      break;
  }
}



void Protein::sproutWholeSidechain(int i) 
{
  for (int n = 0; n < get_n_chi(resKey(i)); n++)
    sproutSidechain(i, n);
}


  
  
void Protein::buildAllHydrogens(int i) 
{
  if (resKey(i) == XXX) return;
  
  Vector3d ca_n = atom(i,N).pos - atom(i,CA).pos;
  Vector3d ca_c = atom(i,C).pos - atom(i,CA).pos;
  Vector3d v    = ca_n.scaleTo(1.0) + ca_c.scaleTo(1.0);
  Vector3d w    = ca_n.scaleTo(1.0) - ca_c.scaleTo(1.0);

  // project amide H atoms
  if ( (resKey(i) == NME) && !hasAtom(i,H) ) 
  {
    Vector3d c_n  = atom(i, N).pos - atom(i-1, C).pos;
    Vector3d n_ca = atom(i, CA).pos - atom(i, N).pos;
    Vector3d axis = cross(n_ca, c_n);
    Vector3d n_h  = n_ca.scaleTo(N_H).rotate(CA_N_H, axis);
    createAtom(i, H, atom(i,N).pos + n_h);
  }
  else if ( (resKey(i) != ACE) && (resKey(i) != PRO) )
  {
    sproutAtom(i, H, N_H, CA_N_H, 180 + phi(i), C, CA, N);
  }
  
  // project aliphatic Ha atoms
  Vector3d ca_ha = v.rotated(-125.5, w).scaleTo(C_H);
  if ( (resKey(i) != ACE) && (resKey(i) != NME) && (!hasAtom(i,HA1)) )
    createAtom(i, HA1, atom(i,CA).pos + ca_ha);

  Vector3d ca_cb = v.rotated(125.5, w);
  if (resKey(i) == GLY)
  {
    if (!hasAtom(i,HA2))
      createAtom(i, HA2, atom(i,CA).pos + ca_cb.scaleTo(C_H));
  }
  else if (resKey(i) == PRO) 
  {
    sproutAtom(i, HB1,  C_H, C_C_H,  -6.11 + 120,  N, CA, CB);
    sproutAtom(i, HB2,  C_H, C_C_H,  -6.11 + 240,  N, CA, CB);

    sproutAtom(i, HG11, C_H, C_C_H,  120, CA, CB, CG1);

    sproutAtom(i, HG12, C_H, C_C_H,  240, CA, CB, CG1);

    sproutAtom(i, HD11, C_H, C_C_H,  120, CA,  N, CD1);
    sproutAtom(i, HD12, C_H, C_C_H,  240, CA,  N, CD1);
  }
  else if (resKey(i) == ALA) 
  {
    sproutAtom(i, HB1, C_H, C_C_H,  60, N, CA, CB);
    sproutAtom(i, HB2, C_H, C_C_H, 180, N, CA, CB);
    sproutAtom(i, HB3, C_H, C_C_H, 300, N, CA, CB);
  }
  else if (resKey(i) == ACE) 
  {
    sproutAtom(i, H1, C_H, C_C_H,  60, O, C, CA);
    sproutAtom(i, H2, C_H, C_C_H, 180, O, C, CA);
    sproutAtom(i, H3, C_H, C_C_H, 300, O, C, CA);
  }
  else if (resKey(i) == NME) 
  {
    sproutAtom(i, H1, C_H, C_C_H,  60 + 60, H, N, CA);
    sproutAtom(i, H2, C_H, C_C_H, 180 + 60, H, N, CA);
    sproutAtom(i, H3, C_H, C_C_H, 300 + 60, H, N, CA);
  }
  
 
  // chi1: chi[0]
  switch (resKey(i)) 
  {
    case ILE:
      sproutAtom(i,  HB1, C_H, C_C_H, chi(i, 0) + 120,  N, CA, CB);
      break;
    case SER:
      sproutAtom(i,  HB1, C_H, C_C_H, chi(i, 0) + 120,  N, CA, CB);
      sproutAtom(i,  HB2, C_H, C_C_H, chi(i, 0) + 240,  N, CA, CB);
      sproutAtom(i, HG11, O_H, 110.0,                 180, CA, CB, OG1);
      break;
    case THR:
      sproutAtom(i,  HB1,   C_H, C_C_H, chi(i, 0) + 120,  N, CA, CB);
      sproutAtom(i, HG11, O_H, 110.0, 180,                 CA, CB, OG1);
      break;
    case VAL:
      sproutAtom(i,  HB1, C_H, C_C_H, chi(i, 0) + 240,  N, CA, CB);
      sproutAtom(i, HG11, C_H, C_C_H,  60, CA, CB, CG1);
      sproutAtom(i, HG12, C_H, C_C_H, 180, CA, CB, CG1);
      sproutAtom(i, HG13, C_H, C_C_H, 300, CA, CB, CG1);
      break;
  }

  switch (resKey(i)) 
  {
    case ASP:
    case ASN:
    case HIS:
    case LEU:
    case GLN:
    case GLU:
    case LYS:
    case ARG:
    case TRP:
    case TYR:
    case PHE:
    case MET:
    case CYS:
      sproutAtom(i,  HB1, C_H, C_C_H, chi(i, 0) + 120,  N, CA, CB);
      sproutAtom(i,  HB2, C_H, C_C_H, chi(i, 0) + 240,  N, CA, CB);
      break;
    case THR:
    case ILE:
    case VAL:
      sproutAtom(i, HG21, C_H, C_C_H,  60, CA, CB, CG2);
      sproutAtom(i, HG22, C_H, C_C_H, 180, CA, CB, CG2);
      sproutAtom(i, HG23, C_H, C_C_H, 300, CA, CB, CG2);
      break;
  }

  // chi2: chi[1]
  switch (resKey(i)) 
  {
    case ASN:
      sproutAtom(i, HD21, N_H, 120.0,    0, CB, CG1, ND2);
      sproutAtom(i, HD22, N_H, 120.0,  180, CB, CG1, ND2);
      break;
      break;
    case ILE:
      sproutAtom(i, HG11, C_H, C_C_H, chi(i, 1) + 120, CA, CB, CG1);
      sproutAtom(i, HG12, C_H, C_C_H, chi(i, 1) + 240, CA, CB, CG1);

      sproutAtom(i, HD11, C_H, C_C_H,   60, CB, CG1, CD1);
      sproutAtom(i, HD12, C_H, C_C_H,  180, CB, CG1, CD1);

      sproutAtom(i, HD13, C_H, C_C_H,  300, CB, CG1, CD1);
      break;
    case LEU:
      sproutAtom(i, HG11, C_H, C_C_H, chi(i, 1) + 240, CA, CB, CG1);

      sproutAtom(i, HD11, C_H, C_C_H,   60, CB, CG1, CD1);
      sproutAtom(i, HD12, C_H, C_C_H,  180, CB, CG1, CD1);
      sproutAtom(i, HD13, C_H, C_C_H,  300, CB, CG1, CD1);

      sproutAtom(i, HD21, C_H, C_C_H,   60, CB, CG1, CD2);
      sproutAtom(i, HD22, C_H, C_C_H,  180, CB, CG1, CD2);
      sproutAtom(i, HD23, C_H, C_C_H,  300, CB, CG1, CD2);
      break;
    case GLN:
    case GLU:
    case LYS:
    case ARG:
      sproutAtom(i, HG11, C_H, C_C_H, chi(i, 1) + 120, CA, CB, CG1);
      sproutAtom(i, HG12, C_H, C_C_H, chi(i, 1) + 240, CA, CB, CG1);
      break;
    case MET:
      sproutAtom(i, HG11, C_H, C_C_H, chi(i, 1) + 120, CA, CB, CG1);
      sproutAtom(i, HG12, C_H, C_C_H, chi(i, 1) + 240, CA, CB, CG1);

      break;
    case TRP:
      sproutAtom(i, HD11, C_H,  124.3,   0,   CB, CG1, CD1);
      sproutAtom(i, HE11, N_H,  126.8, 180,  CG1, CD1, NE1);
      sproutAtom(i, HZ21, C_H,  125.8, 180,  CD2, CE2, CZ2);
      sproutAtom(i, HH21, C_H,  125.8, 180,  CE2, CZ2, CH2);
      sproutAtom(i, HZ31, C_H,  125.8, 180,  CD2, CE3, CZ3);
      sproutAtom(i, HE31, C_H,  125.8,   0,  CG1, CD2, CE3);
      break;
    case HIS:
      sproutAtom(i, HD11, C_H,  125.8,     0,   CB, CG1, ND1);
      sproutAtom(i, HE11, N_H,  125.8,   180,  CG1, ND1, CE1);
      sproutAtom(i, HD21, C_H,  125.8,     0,   CB, CG1, CD2);
      sproutAtom(i, HE21, N_H,  125.8,   180,  ND1, CE1, NE2);
      break; 
    case TYR:
    case PHE:
      sproutAtom(  i, HD11, C_H, 120.0,    0,  CB, CG1, CD1);
      sproutAtom(  i, HD21, C_H, 120.0,    0,  CB, CG1, CD2);

      sproutAtom(  i, HE11, C_H, 120.0,  180, CG1, CD1, CE1);
      sproutAtom(  i, HE21, C_H, 120.0,  180, CG1, CD2, CE2);

      if (resKey(i) == PHE)
        sproutAtom(i, HZ11, C_H, 120.0,  180, CD1, CE1, CZ1);
      else {
        sproutAtom(i, HH11, C_H, 110.0,  180, CE1, CZ1, OH1);
      }

      break;
  }

  // chi3: chi[2]
  switch (resKey(i)) 
  {
    case GLN:
      sproutAtom(i, HE21, N_H, 120.0,    0, CG1, CD1, NE2);
      sproutAtom(i, HE22, N_H, 120.0,  180, CG1, CD1, NE2);
      break;
    case ARG:
      sproutAtom(i, HD11, C_H, C_C_H, chi(i, 2) + 120, CB, CG1, CD1);
      sproutAtom(i, HD12, C_H, C_C_H, chi(i, 2) + 240, CB, CG1, CD1);
      break;
    case LYS:

      sproutAtom(i, HD11, C_H, C_C_H, chi(i, 2) + 120, CB, CG1, CD1);
      sproutAtom(i, HD12, C_H, C_C_H, chi(i, 2) + 240, CB, CG1, CD1);
      break;
    case MET:
      sproutAtom(i, HE11, C_H, C_C_H,  60, CG1, SD1, CE1);
      sproutAtom(i, HE12, C_H, C_C_H, 180, CG1, SD1, CE1);
      sproutAtom(i, HE13, C_H, C_C_H, 300, CG1, SD1, CE1);
      break;
  }

  // chi4: chi[3]
  switch (resKey(i)) 
  {
    case ARG:
      sproutAtom(i, HE11, C_H, 240.0, chi(i, 3), CG1, CD1, NE1);

      sproutAtom(i, HH11, N_H, 120.0,   0,  NE1, CZ1, NH1);
      sproutAtom(i, HH12, N_H, 120.0, 180,  NE1, CZ1, NH1);

      sproutAtom(i, HH21, N_H, 120.0,   0,  NE1, CZ1, NH2);
      sproutAtom(i, HH22, N_H, 120.0, 180,  NE1, CZ1, NH2);

      break;
    case LYS:
      sproutAtom(i, HE11, C_H, C_C_H, chi(i, 3) + 120, CG1, CD1, CE1);
      sproutAtom(i, HE12, C_H, C_C_H, chi(i, 3) + 240, CG1, CD1, CE1);

      sproutAtom(i, HZ11, C_H, C_C_H,   60, CD1, CE1, NZ1);
      sproutAtom(i, HZ12, C_H, C_C_H,  180, CD1, CE1, NZ1);
      sproutAtom(i, HZ13, C_H, C_C_H,  300, CD1, CE1, NZ1);
      break;
  }
}



// --------- dihedral angle rotations



void Protein::extractProlineChiAngles(int i)
{
  if (resKey(i) != PRO) return;
  
  chi(i, 0) = pos_dihedral(atom(i, N  ).pos, atom(i, CA ).pos, atom(i, CB ).pos, atom(i, CG1).pos );
  chi(i, 1) = pos_dihedral(atom(i, CA ).pos, atom(i, CB ).pos, atom(i, CG1).pos, atom(i, CD1).pos );
  chi(i, 2) = pos_dihedral(atom(i, CB ).pos, atom(i, CG1).pos, atom(i, CD1).pos, atom(i, CA ).pos );
  chi(i, 3) = pos_dihedral(atom(i, CG1).pos, atom(i, CD1).pos, atom(i, N  ).pos, atom(i, CA ).pos );
  
}



void Protein::extractProlineParam(int i)
{
  if (resKey(i) != PRO) return;
  if (chi(i, 1) > 0)
    proPucker(i) = +1.0;
  else
    proPucker(i) = -1.0;
  proTheta(i) = DEG2RAD * pos_angle(atom(i, CG1).pos, atom(i, CA).pos, atom(i, CD1).pos);
  proXi(i)    = DEG2RAD * pos_angle(atom(i, N  ).pos, atom(i, CA).pos, atom(i, CD1).pos);
  proEta(i)   = DEG2RAD * pos_angle(atom(i, CB ).pos, atom(i, CA).pos, atom(i, CG1).pos);
  proAlpha(i) = DEG2RAD * pos_angle(atom(i, N  ).pos, atom(i, CA).pos, atom(i, CB ).pos);
}



bool Protein::projectProlineRing(int i, double newPhi)
{
  if (resKey(i) != PRO) 
    return false;
  
  double old_chi1    = chi(i, 0);
  double old_chi5    = pos_dihedral(atom(i, CB).pos, atom(i,  CA).pos, 
                                    atom(i,  N).pos, atom(i, CD1).pos );
  double phi_to_chi5 = phi(i) - old_chi5;
  double chi5        = newPhi - phi_to_chi5;
  double tau       = DEG2RAD * chi5;

  double theta     = proTheta(i);  // CG-CA-CD
  double xi        = proXi(i);     // N-CA-CD
  double eta       = proEta(i);    // CB-CA-CG
  double alpha     = proAlpha(i);  // N-CA-CB

  double a  =  - cos(alpha) * sin(xi ) * sin(eta) * cos(tau) 
               + sin(alpha) * sin(xi ) * cos(eta);
  double b  =  - sin(xi   ) * sin(eta) * sin(tau);
  double c  =    cos(theta) 
               - cos(alpha) * cos(xi ) * cos(eta) 
               - sin(alpha) * cos(xi ) * sin(eta) * cos(tau);  

  double denom = sqrt(a*a + b*b);
  double z  = c / denom;
  
  if ( (z < -1.0) || (z > 1.0) )
    return false;
     
  double sigma1 = acos( z );
  double sigma0;
  if (b >= 0)
    sigma0 = + acos(a / denom);
  else 
    sigma0 = - acos(a / denom);
  double sigma = sigma0 - proPucker(i) * sigma1;

  double chi1       = RAD2DEG * sigma;
  double d_chi1     = chi1 - old_chi1;
  atom(i, CG1 ).pos = rotate_pos(-d_chi1, atom(i, CG1).pos, atom(i, CA).pos, atom(i, CB).pos );

  double d_chi5     = chi5 - old_chi5;
  atom(i, CD1 ).pos = rotate_pos(-d_chi5, atom(i, CD1).pos, atom(i, CA).pos, atom(i, N).pos );

  atom(i, HD11).pos = project_tetr_up_pos(  1.1, atom(i,N).pos, atom(i,CD1).pos, atom(i,CG1).pos); 
  atom(i, HD12).pos = project_tetr_down_pos(1.1, atom(i,N).pos, atom(i,CD1).pos, atom(i,CG1).pos); 
  

  atom(i, HG11).pos = project_tetr_up_pos(  1.1, atom(i,CD1).pos, atom(i,CG1).pos, atom(i,CB).pos); 
  atom(i, HG12).pos = project_tetr_down_pos(1.1, atom(i,CD1).pos, atom(i,CG1).pos, atom(i,CB).pos); 

  atom(i,  HB1).pos = project_tetr_up_pos(  1.1, atom(i,CG1).pos, atom(i,CB).pos, atom(i,CA).pos); 
  atom(i,  HB2).pos = project_tetr_down_pos(1.1, atom(i,CG1).pos, atom(i,CB).pos, atom(i,CA).pos); 

  extractProlineChiAngles(i);
  
  return true;
}



void Protein::setProPucker(int i, double pucker) 
{
  if (pucker < 0.0)
    proPucker(i) = -1.0;
  else
    proPucker(i) = 1.0;
  projectProlineRing(i, phi(i));
}



void Protein::setBackbonePhi(int i, double newPhi) 
{
  if ( (resKey(i) == ACE) || (resKey(i) == NME) ) 
    return;

  Vector3d center  = atom(i,N).pos;
  Vector3d axis    = atom(i,N).pos - atom(i,CA).pos;
  double dPhi      = newPhi - phi(i);

  Matrix3d rotation = RotationMatrix(axis, dPhi, center);

  if (resKey(i) != PRO)
  {  
    atom(i,H).pos.transform(rotation);
    if (hasAtom(i, H2)) atom(i,H2).pos.transform(rotation);
    if (hasAtom(i, H3)) atom(i,H3).pos.transform(rotation);
  }

  for (int seg=0; seg < i; seg++)
    res(seg).transform(rotation);

  phi(i) = newPhi;
}



void Protein::setPhi(int i, double newPhi) 
{
  if (resKey(i) == PRO)
  {
    if (projectProlineRing(i, newPhi))
      setBackbonePhi(i, newPhi);
  }
  else
    setBackbonePhi(i, newPhi);
}



void Protein::setPsi(int i, double newPsi) 
{
  if ( (resKey(i) == ACE) || (resKey(i) == NME) ) return;

  Vector3d center  = atom(i,C).pos;
  Vector3d axis    = atom(i,C).pos - atom(i,CA).pos;
  double d_psi      = newPsi - psi(i);

  Matrix3d rotMatrix = RotationMatrix(axis, d_psi, center);

  if (i < nRes()-1) 
    for (int seg=i+1; seg < nRes(); seg++)
       res(seg).transform(rotMatrix);

  atom(i,O).pos.transform(rotMatrix);
  if (hasAtom(i, OXT)) atom(i,OXT).pos.transform(rotMatrix);

  psi(i) = newPsi;
}



void Protein::setChi(int i, int n, double newChi) 
{
  if (resKey(i) == PRO) 
    return;

  int pivot_atom_key  = res_chi_tetrad[resKey(i)][n][1];
  int center_atom_key = res_chi_tetrad[resKey(i)][n][2];

  Vector3d pivot    = atom(i, pivot_atom_key).pos;
  Vector3d center   = atom(i, center_atom_key).pos;
  Vector3d axis     = center - pivot;
  double   dChi     = newChi - chi(i, n);

  Matrix3d rotMatrix = RotationMatrix(axis, dChi, center);

  for (int j = 0; j < res(i).nAtom(); j++) 
  {
    if (res(i).atomIters[j]->key > center_atom_key)
      res(i).atomIters[j]->pos.transform(rotMatrix);
  }

  chi(i, n) = newChi;
}




void Protein::blockTerminals()
{
  if (nRes() == 0)
    return;
    
  if (resKey(0) != ACE)
  {
    unblockNTerminal();
    createRes(0, ACE);
    
  }

  if (resKey(nRes()-1) != NME)
  {
    unblockCTerminal();
    appendRes(NME);    
  } 
}


void Protein::chargeTerminals()
{
  if (resKey(0) == ACE)
    unblockNTerminal(); 
  sproutAtom(0, H2, N_H, CA_N_H, 180 + phi(0) + 120, C, CA, N);
  sproutAtom(0, H3, N_H, CA_N_H, 180 + phi(0) + 240, C, CA, N);

  int last_res = nRes() - 1;
  if (resKey(last_res) == NME)
    unblockCTerminal();
  sproutAtom(last_res, OXT, C_O, CA_C_O, 180 + psi(last_res) + 180, N, CA, C);
  sproutAtom(last_res, OXT, C_O, CA_C_O, 180 + psi(last_res) + 180, N, CA, C);
}



void Protein::unblockNTerminal()
{
  if (resKey(0) == ACE)
    eraseRes(0);
  else
  {
    eraseAtom(0, H2);
    eraseAtom(0, H3);
  }
}


void Protein::unblockCTerminal()
{
  int last_res = nRes() - 1;
  if (resKey(last_res) == NME) 
    eraseRes(last_res);
  else
    eraseAtom(last_res, OXT);
}


